#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the 'Software'), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

import datetime
import pprint
import time

import xmltodict

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxPluginExec/plugins/EDPluginGroupISPyB-v1.4/plugins/
#     EDPluginISPyBRetrieveDataCollectionv1_4.py


from suds.client import Client
from suds.transport.http import HttpAuthenticated

import os
import gzip
import pathlib

from edna2.utils import UtilsPath
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging
from edna2.utils import UtilsIspyb

from edna2.tasks.AbstractTask import AbstractTask

logger = UtilsLogging.getLogger()


class ISPyBRetrieveDataCollection(AbstractTask):
    def run(self, inData):
        dictConfig = UtilsConfig.getTaskConfig("ISPyB")
        username = dictConfig["username"]
        password = dictConfig["password"]
        httpAuthenticated = HttpAuthenticated(username=username, password=password)
        wdsl = dictConfig["ispyb_ws_url"] + "/ispybWS/ToolsForCollectionWebService?wsdl"
        client = Client(wdsl, transport=httpAuthenticated, cache=None)
        if "image" in inData:
            path = pathlib.Path(inData["image"])
            indir = path.parent.as_posix()
            infilename = path.name
            dataCollection = (
                client.service.findDataCollectionFromFileLocationAndFileName(
                    indir, infilename
                )
            )
        elif "dataCollectionId" in inData:
            dataCollectionId = inData["dataCollectionId"]
            dataCollection = client.service.findDataCollection(dataCollectionId)
        else:
            errorMessage = "Neither image nor data collection id given as input!"
            logger.error(errorMessage)
            raise BaseException(errorMessage)
        if dataCollection is not None:
            outData = Client.dict(dataCollection)
        else:
            outData = {}
        return outData


class GetListAutoprocIntegration(AbstractTask):
    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "proposal": {"type": "string"},
                "dataCollectionId": {"type": "integer"},
            },
        }

    # def getOutDataSchema(self):
    #     return {
    #         "type": "array",
    #         "items": {
    #             "type": "object",
    #             "properties": {
    #                 "AutoProcIntegration_autoProcIntegrationId": {"type": "integer"}
    #             }
    #         }
    #     }

    def run(self, inData):
        # urlExtISPyB, token, proposal, dataCollectionId
        token = inData["token"]
        proposal = inData["proposal"]
        dataCollectionId = inData["dataCollectionId"]
        dictConfig = UtilsConfig.getTaskConfig("ISPyB")
        restUrl = dictConfig["ispyb_ws_url"] + "/rest"
        ispybWebServiceURL = os.path.join(
            restUrl,
            token,
            "proposal",
            str(proposal),
            "mx",
            "autoprocintegration",
            "datacollection",
            str(dataCollectionId),
            "view",
        )
        dataFromUrl = UtilsIspyb.getDataFromURL(ispybWebServiceURL)
        outData = {}
        if dataFromUrl["statusCode"] == 200:
            outData["autoprocIntegration"] = dataFromUrl["data"]
        else:
            outData["error"] = dataFromUrl
        return outData


class GetListAutoprocAttachment(AbstractTask):
    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "proposal": {"type": "string"},
                "autoProcProgramId": {"type": "integer"},
            },
        }

    # def getOutDataSchema(self):
    #     return {
    #         "type": "array",
    #         "items": {
    #             "type": "object",
    #             "properties": {
    #                 "AutoProcIntegration_autoProcIntegrationId": {"type": "integer"}
    #             }
    #         }
    #     }

    def run(self, inData):
        # urlExtISPyB, token, proposal, autoProcProgramId
        token = inData["token"]
        proposal = inData["proposal"]
        autoProcProgramId = inData["autoProcProgramId"]
        dictConfig = UtilsConfig.getTaskConfig("ISPyB")
        restUrl = dictConfig["ispyb_ws_url"] + "/rest"
        ispybWebServiceURL = os.path.join(
            restUrl,
            token,
            "proposal",
            str(proposal),
            "mx",
            "autoprocintegration",
            "attachment",
            "autoprocprogramid",
            str(autoProcProgramId),
            "list",
        )
        dataFromUrl = UtilsIspyb.getDataFromURL(ispybWebServiceURL)
        outData = {}
        if dataFromUrl["statusCode"] == 200:
            outData["autoprocAttachment"] = dataFromUrl["data"]
        else:
            outData["error"] = dataFromUrl
        return outData


class GetListAutoprocessingResults(AbstractTask):
    """
    This task receives a list of data collection IDs and returns a list
    of dictionaries with all the auto-processing results and file attachments
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "proposal": {"type": "string"},
                "dataCollectionId": {
                    "type": "array",
                    "items": {
                        "type": "integer",
                    },
                },
            },
        }

    # def getOutDataSchema(self):
    #     return {
    #         "type": "object",
    #         "required": ["dataForMerge"],
    #         "properties": {
    #             "dataForMerge": {
    #                 "type": "object",
    #                 "items": {
    #                     "type": "object",
    #                     "properties": {
    #                         "spaceGroup": {"type": "string"}
    #                     }
    #                 }
    #             }
    #         }
    #     }

    def run(self, inData):
        urlError = None
        token = inData["token"]
        proposal = inData["proposal"]
        listDataCollectionId = inData["dataCollectionId"]
        dictForMerge = {}
        dictForMerge["dataCollection"] = []
        for dataCollectionId in listDataCollectionId:
            dictDataCollection = {"dataCollectionId": dataCollectionId}
            inDataGetListIntegration = {
                "token": token,
                "proposal": proposal,
                "dataCollectionId": dataCollectionId,
            }
            getListAutoprocIntegration = GetListAutoprocIntegration(
                inData=inDataGetListIntegration
            )
            getListAutoprocIntegration.setPersistInOutData(False)
            getListAutoprocIntegration.execute()
            resultAutoprocIntegration = getListAutoprocIntegration.outData
            if "error" in resultAutoprocIntegration:
                urlError = resultAutoprocIntegration["error"]
                break
            else:
                listAutoprocIntegration = resultAutoprocIntegration[
                    "autoprocIntegration"
                ]
                # Get v_datacollection_summary_phasing_autoProcProgramId
                for autoprocIntegration in listAutoprocIntegration:
                    if (
                        "v_datacollection_summary_phasing_autoProcProgramId"
                        in autoprocIntegration
                    ):
                        autoProcProgramId = autoprocIntegration[
                            "v_datacollection_summary_phasing_autoProcProgramId"
                        ]
                        inDataGetListAttachment = {
                            "token": token,
                            "proposal": proposal,
                            "autoProcProgramId": autoProcProgramId,
                        }
                        getListAutoprocAttachment = GetListAutoprocAttachment(
                            inData=inDataGetListAttachment
                        )
                        getListAutoprocAttachment.setPersistInOutData(False)
                        getListAutoprocAttachment.execute()
                        resultAutoprocAttachment = getListAutoprocAttachment.outData
                        if "error" in resultAutoprocAttachment:
                            urlError = resultAutoprocAttachment["error"]
                        else:
                            autoprocIntegration[
                                "autoprocAttachment"
                            ] = resultAutoprocAttachment["autoprocAttachment"]
                    dictDataCollection["autoprocIntegration"] = listAutoprocIntegration
            dictForMerge["dataCollection"].append(dictDataCollection)
            # dictForMerge[dataCollectionId] = dictDataCollection
        if urlError is None:
            outData = dictForMerge
        else:
            outData = {"error": urlError}
        return outData


class RetrieveAttachmentFiles(AbstractTask):
    """
    This task receives a list of data collection IDs and returns a list
    of dictionaries with all the auto-processing results and file attachments
    """

    # def getInDataSchema(self):
    #     return {
    #         "type": "object",
    #         "properties": {
    #             "token": {"type": "string"},
    #             "proposal": {"type": "string"},
    #             "dataCollectionId": {
    #                 "type": "array",
    #                 "items": {
    #                     "type": "integer",
    #                 }
    #             }
    #         }
    #     }

    # def getOutDataSchema(self):
    #     return {
    #         "type": "object",
    #         "required": ["dataForMerge"],
    #         "properties": {
    #             "dataForMerge": {
    #                 "type": "object",
    #                 "items": {
    #                     "type": "object",
    #                     "properties": {
    #                         "spaceGroup": {"type": "string"}
    #                     }
    #                 }
    #             }
    #         }
    #     }

    def run(self, inData):
        urlError = None
        listPath = []
        token = inData["token"]
        proposal = inData["proposal"]
        listAttachment = inData["attachment"]
        dictConfig = UtilsConfig.getTaskConfig("ISPyB")
        restUrl = dictConfig["ispyb_ws_url"] + "/rest"
        # proposal/MX2112/mx/autoprocintegration/autoprocattachmentid/21494689/get
        for dictAttachment in listAttachment:
            attachmentId = dictAttachment["id"]
            fileName = dictAttachment["fileName"]
            ispybWebServiceURL = os.path.join(
                restUrl,
                token,
                "proposal",
                str(proposal),
                "mx",
                "autoprocintegration",
                "autoprocattachmentid",
                str(attachmentId),
                "get",
            )
            rawDataFromUrl = UtilsIspyb.getRawDataFromURL(ispybWebServiceURL)
            if rawDataFromUrl["statusCode"] == 200:
                rawData = rawDataFromUrl["content"]
                if fileName.endswith(".gz"):
                    rawData = gzip.decompress(rawData)
                    fileName = fileName.split(".gz")[0]
                with open(fileName, "wb") as f:
                    f.write(rawData)
                listPath.append(str(self.getWorkingDirectory() / fileName))
            else:
                urlError = rawDataFromUrl
        if urlError is None:
            outData = {"filePath": listPath}
        else:
            outData = {"error": urlError}
        return outData


class ISPyBFindDetectorByParam(AbstractTask):
    def run(self, inData):
        dictConfig = UtilsConfig.getTaskConfig("ISPyB")
        username = dictConfig["username"]
        password = dictConfig["password"]
        httpAuthenticated = HttpAuthenticated(username=username, password=password)
        wdsl = dictConfig["ispyb_ws_url"] + "/ispybWS/ToolsForCollectionWebService?wsdl"
        client = Client(wdsl, transport=httpAuthenticated, cache=None)
        manufacturer = inData["manufacturer"]
        model = inData["model"]
        mode = inData["mode"]
        detector = client.service.findDetectorByParam("", manufacturer, model, mode)
        if detector is not None:
            outData = Client.dict(detector)
        else:
            outData = {}
        return outData


class UploadGPhLResultsToISPyB(AbstractTask):
    def run(self, in_data):
        list_id = []
        # List of data collection ids
        list_data_collection_id = in_data["dataCollectionId"]
        # Load XML file
        xml_path = in_data["autoPROCXML"]
        with open(xml_path) as f:
            xml_string = f.read()
        auto_proc_dict = xmltodict.parse(xml_string)
        pprint.pprint(auto_proc_dict)
        auto_proc_container = auto_proc_dict["AutoProcContainer"]
        auto_proc = auto_proc_container["AutoProc"]
        if type(auto_proc) == list:
            list_auto_proc = auto_proc
        else:
            list_auto_proc = [auto_proc]
        auto_proc_program_container = auto_proc_container["AutoProcProgramContainer"]
        auto_proc_scaling_container = auto_proc_container["AutoProcScalingContainer"]
        if type(auto_proc_scaling_container) == list:
            list_auto_proc_scaling_container = auto_proc_scaling_container
        else:
            list_auto_proc_scaling_container = [auto_proc_scaling_container]
        for index, auto_proc_scaling_container in enumerate(list_auto_proc_scaling_container):
            auto_proc = list_auto_proc[index]
            data_collection_id = list_data_collection_id[index]
            auto_proc_scaling = auto_proc_scaling_container["AutoProcScaling"]
            auto_proc_scaling_statistics = auto_proc_scaling_container[
                "AutoProcScalingStatistics"
            ]
            auto_proc_integration_container = auto_proc_scaling_container[
                "AutoProcIntegrationContainer"
            ]
            auto_proc_integration = auto_proc_integration_container["AutoProcIntegration"]
            # image =
            # 1. Create AutoProcProgram entry
            auto_proc_program = auto_proc_program_container["AutoProcProgram"]
            auto_proc_program_id = self.createAutoProcProgram(auto_proc_program)
            # 2. Upload all AutoProcProgramAttachments
            list_attachment = auto_proc_program_container["AutoProcProgramAttachment"]
            list_attachment_id = self.uploadAttachments(
                auto_proc_program_id, list_attachment
            )
            # 3. Create AutoProcIntegration entry
            auto_proc_integration_id = self.createAutoProcIntegration(
                auto_proc_program_id=auto_proc_program_id,
                data_collection_id=data_collection_id,
                auto_proc_integration=auto_proc_integration,
            )
            # 4. Create AutoProc entry
            auto_proc_id = self.createAutoProc(
                auto_proc_program_id=auto_proc_program_id, auto_proc=auto_proc
            )
            # 5. Create scaling entry
            auto_proc_scaling_id = self.createAutoProcScaling(
                auto_proc_id=auto_proc_id, auto_proc_scaling=auto_proc_scaling
            )
            # 6. Create scaling has int entry
            auto_proc_scaling_has_int_id = self.createAutoProcScaling_has_int(
                auto_proc_integration_id=auto_proc_integration_id,
                auto_proc_scaling_id=auto_proc_scaling_id
            )
            # 6. Create scaling statistics entries
            list_statistics_id = self.uploadStatistics(
                auto_proc_scaling_id=auto_proc_scaling_id,
                auto_proc_scaling_statistics=auto_proc_scaling_statistics,
            )
            scaling_container_id = {
                "data_collection_id": data_collection_id,
                "auto_proc_id": auto_proc_id,
                "auto_proc_program_id": auto_proc_program_id,
                "auto_proc_integration_id": auto_proc_integration_id,
                "list_attachment_id": list_attachment_id,
                "auto_proc_scaling_id": auto_proc_scaling_id,
                "auto_proc_scaling_has_int_id": auto_proc_scaling_has_int_id,
                "list_statistics_id": list_statistics_id,
            }
            list_id.append(scaling_container_id)
        out_data = {
            "ids": list_id
        }
        pprint.pprint(out_data)
        return out_data

    def createAutoProcProgram(self, auto_proc_program):
        auto_proc_program_id = UtilsIspyb.storeOrUpdateAutoProcProgram(
            programs=self.check_length(auto_proc_program["processingPrograms"]),
            commandline=self.check_length(auto_proc_program["processingCommandLine"]),
            start_time=self.get_time(auto_proc_program["processingStartTime"]),
            end_time=self.get_time(auto_proc_program["processingEndTime"]),
            environment=self.check_length(auto_proc_program["processingEnvironment"]),
            message=self.check_length(auto_proc_program["processingMessage"]),
            status=auto_proc_program["processingStatus"],
        )
        return auto_proc_program_id

    def uploadAttachments(self, auto_proc_program_id, list_attachments):
        list_attachment_id = []
        for dict_attachment in list_attachments:
            file_name = dict_attachment["fileName"]
            file_dir = pathlib.Path(dict_attachment["filePath"])
            file_path = file_dir / file_name
            if file_path.exists():
                file_type = dict_attachment["fileType"]
                pyarch_dir = UtilsPath.createPyarchFilePath(file_dir)
                pyarch_dir.mkdir(mode=0o755, exist_ok=True, parents=True)
                pyarch_path = pyarch_dir / file_name
                UtilsPath.systemCopyFile(file_path, pyarch_path)
                attachment_id = UtilsIspyb.storeOrUpdateAutoProcProgramAttachment(
                    auto_proc_program_id=auto_proc_program_id,
                    file_path=pyarch_path,
                    file_type=file_type,
                )
                list_attachment_id.append(attachment_id)
            else:
                logger.warning(f"File path {file_path} doesn't exists")
        return list_attachment_id

    def createAutoProcIntegration(
        self, auto_proc_program_id, data_collection_id, auto_proc_integration
    ):
        auto_proc_integration_id = UtilsIspyb.storeOrUpdateAutoProcIntegration(
            auto_proc_program_id=auto_proc_program_id,
            dataCollection_id=data_collection_id,
            start_image_number=auto_proc_integration.get("startImageNumber", None),
            end_image_number=auto_proc_integration.get("endImageNumber", None),
            refined_detector_distance=auto_proc_integration.get(
                "refinedDetectorDistance", None
            ),
            refined_x_beam=auto_proc_integration.get("refinedXBeam", None),
            refined_y_beam=auto_proc_integration.get("refinedYBeam", None),
            rotation_axis_x=auto_proc_integration.get("rotationAxisX", None),
            rotation_axis_y=auto_proc_integration.get("rotationAxisY", None),
            rotation_axis_z=auto_proc_integration.get("rotationAxisZ", None),
            beam_vector_x=auto_proc_integration.get("beamVectorX", None),
            beam_vector_y=auto_proc_integration.get("beamVectorY", None),
            beam_vector_z=auto_proc_integration.get("beamVectorZ", None),
            cell_a=auto_proc_integration.get("cell_a", None),
            cell_b=auto_proc_integration.get("cell_b", None),
            cell_c=auto_proc_integration.get("cell_c", None),
            cell_alpha=auto_proc_integration.get("cell_alpha", None),
            cell_beta=auto_proc_integration.get("cell_beta", None),
            cell_gamma=auto_proc_integration.get("cell_gamma", None),
            record_time_stamp=None,
            anomalous=1,
        )
        return auto_proc_integration_id

    def createAutoProc(self, auto_proc_program_id, auto_proc):
        auto_proc_id = UtilsIspyb.storeOrUpdateAutoProc(
            auto_proc_program_id=auto_proc_program_id,
            space_group=auto_proc["spaceGroup"],
            refined_cell_a=auto_proc["refinedCell_a"],
            refined_cell_b=auto_proc["refinedCell_b"],
            refined_cell_c=auto_proc["refinedCell_c"],
            refined_cell_alpha=auto_proc["refinedCell_alpha"],
            refined_cell_beta=auto_proc["refinedCell_beta"],
            refined_cell_gamma=auto_proc["refinedCell_gamma"],
        )
        return auto_proc_id

    def createAutoProcScaling(self, auto_proc_id, auto_proc_scaling):
        auto_proc_scaling_id = UtilsIspyb.storeOrUpdateAutoProcScaling(
            auto_proc_id=auto_proc_id
        )
        return auto_proc_scaling_id

    def createAutoProcScaling_has_int(self, auto_proc_integration_id, auto_proc_scaling_id):
        auto_proc_scaling_has_int_id = UtilsIspyb.storeOrUpdateAutoProcScalingHasInt(
            auto_proc_integration_id=auto_proc_integration_id,
            auto_proc_scaling_id=auto_proc_scaling_id
        )
        return auto_proc_scaling_has_int_id


    def uploadStatistics(self, auto_proc_scaling_id, auto_proc_scaling_statistics):
        list_statistics_id = []
        for statistics in auto_proc_scaling_statistics:
            auto_proc_scaling_statistics_id = (
                UtilsIspyb.storeOrUpdateAutoProcScalingStatistics(
                    scaling_statistics_type=statistics["scalingStatisticsType"],
                    resolution_limit_low=statistics["resolutionLimitLow"],
                    resolution_limit_high=statistics["resolutionLimitHigh"],
                    r_merge=statistics["rMerge"],
                    r_meas_within_i_plus_i_minus=statistics["rMeasWithinIPlusIMinus"],
                    r_meas_all_i_plus_i_minus=statistics["rMeasAllIPlusIMinus"],
                    r_pim_within_i_plus_i_minus=statistics["rPimWithinIPlusIMinus"],
                    r_pim_all_i_plus_i_minus=statistics["rPimAllIPlusIMinus"],
                    ntotal_observations=statistics["nTotalObservations"],
                    ntotal_unique_observations=statistics["nTotalUniqueObservations"],
                    mean_iover_sig_i=statistics["meanIOverSigI"],
                    completeness=statistics["completeness"],
                    multiplicity=statistics["multiplicity"],
                    anomalous_completeness=statistics["anomalousCompleteness"],
                    anomalous_multiplicity=statistics["anomalousMultiplicity"],
                    auto_proc_scaling_id=auto_proc_scaling_id,
                    cc_half=statistics["ccHalf"],
                    cc_ano=statistics["ccAnomalous"],
                )
            )
            list_statistics_id.append(auto_proc_scaling_statistics_id)
        return list_statistics_id

    def check_length(self, parameter, max_string_length=255):
        if type(parameter) == str and len(parameter) > max_string_length:
            old_parameter = parameter
            parameter = parameter[0 : max_string_length - 3] + "..."
            logger.warning(
                "String truncated to %d characters for ISPyB! Original string: %s"
                % (max_string_length, old_parameter)
            )
            logger.warning("Truncated string: %s" % parameter)
        return parameter

    def get_time(self, time_value):
        # Fri May 12 08:31:54 CEST 2023
        return datetime.datetime.strptime(time_value, "%a %b %d %H:%M:%S %Z %Y")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import os
import sys
import json
import logging
import pathlib
import jsonschema

import autocryst.saveDozor as sd
from autocryst.Image import ImageHandler as Im
from autocryst import run_crystfel

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsPath
from edna2.utils import UtilsImage
from edna2.utils import UtilsIspyb
from edna2.utils import UtilsLogging

__authors__ = ["S. Basu", "Olof Svensson"]
__license__ = "MIT"
__date__ = "05/07/2019"

logger = UtilsLogging.getLogger()


class ExeCrystFEL(AbstractTask):
    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "listH5FilePath": {"type": "array", "items": {"type": "string"}},
                "doCBFtoH5": {"type": "boolean"},
                "doSubmit": {"type": "boolean"},
                "detectorType": {"type": "string"},
                "batchSize": {"type": "integer"},
                "cbfFileInfo": {
                    "directory": {"type": "string"},
                    "template": {"type": "string"},
                    "startNo": {"type": "integer"},
                    "endNo": {"type": "integer"},
                    "batchSize": {"type": "integer"},
                    "listofImages": {"type": "array", "items": {"type": "string"}},
                },
                "imageQualityIndicators": {
                    "type": "array",
                    "items": {"$ref": self.getSchemaUrl("imageQualityIndicators.json")},
                },
            },
            "oneOf": [{"required": ["listH5FilePath"]}, {"required": ["cbfFileInfo"]}],
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "first_data_path": {"type": "string"},
                "streamfile": {"type": "string"},
                "centering": {"type": "string"},
                "num_indexed_frames": {"type": "integer"},
                "lattice": {"type": "string"},
                "unique_axis": {"type": "string"},
                "unit_cell": {
                    "type": "array",
                    "items": {"type": "number"},
                },
                "point_group": {"type": "string"},
                "space_group": {"type": "string"},
                "resolution_limit": {"type": "number"},
                "average_num_spots": {"type": "number"},
            },
        }

    def run(self, inData):
        # Determine data diretcory
        if "listH5FilePath" in inData:
            first_data_path = inData["listH5FilePath"][0]
        elif "cbfFileInfo" in inData:
            raise RuntimeError("Not yet implemented!")
        else:
            raise RuntimeError("No data source found inData")
        doCBFtoH5 = inData.get("doCBFtoH5", False)

        outData = {}
        if doCBFtoH5:
            dd = sd.Dozor(inData)
            dd.extract_olof_json(inData)

            headerfile = self.getWorkingDirectory() / "headers.json"
            if dd.is_success():
                os.chdir(str(self.getWorkingDirectory()))
                if not headerfile.exists():
                    with open(str(headerfile), "w") as jhead:
                        json.dump(dd.cbfheader, jhead, sort_keys=True, indent=2)
                else:
                    pass

                if dd.stacklength <= 100:
                    dd.create_stack()
                else:
                    dd.mp_stack()

                outData = self.exeIndexing(inData)
            else:
                self.setFailure()
                logger.error(
                    "CrystFEL Task failed due to failure of dozor packing into cxi"
                )
        else:
            os.chdir(self.getWorkingDirectory())
            streampath, results = self.exeIndexing(inData)
            if streampath.exists():
                outData = results
                outData["streamfile"] = str(streampath)
                outData["first_data_path"] = first_data_path
            else:
                self.isFailure()
                logger.error("AutoCryst returned empty stream file")
        return outData

    def exeIndexing(self, inData):
        doCBFtoH5 = inData.get("doCBFtoH5", False)
        doSubmit = inData.get("doSubmit", True)
        in_for_crystfel = dict()

        if "listH5FilePath" in inData.keys():
            tmp = UtilsImage.getPrefix(inData["listH5FilePath"][0])
            in_for_crystfel["prefix"] = tmp
            in_for_crystfel["detectorType"] = inData.get("detectorType", "jungfrau")
            if inData.get("detectorType", "jungfrau") == "eiger":
                in_for_crystfel["prefix"] = tmp.strip("data")
                in_for_crystfel["maxchunksize"] = 10
                FirstImage = tmp.replace("data", "master.h5")
                Image = Im(FirstImage)
                in_for_crystfel["detectorType"] = (
                    Image.imobject.headers["detector_name"][0]
                    + Image.imobject.headers["detector_name"][1]
                )
            else:
                in_for_crystfel["geometry_file"] = inData["geometry_file"]
                in_for_crystfel["threshold"] = inData.get("threshold", "1000")
                in_for_crystfel["peak_radius"] = inData.get("peak_radius", "4,5,10")
                in_for_crystfel["int_radius"] = inData.get("int_radius", "4,5,10")
                in_for_crystfel["min_snr"] = inData.get("min_snr", "5.0")
                in_for_crystfel["min_peaks"] = inData.get("min_peaks", "20")
                in_for_crystfel["max_res"] = inData.get("max_res", "1200")
                in_for_crystfel["partition"] = inData.get("partition", "mx-low")
                in_for_crystfel["unit_cell_file"] = inData.get("unit_cell_file", " ")
                in_for_crystfel["indexing_method"] = inData.get("indexing_method", "xgandalf")
                in_for_crystfel["local_bg_radius"] = inData.get("local_bg_radius", "10")

            in_for_crystfel["suffix"] = UtilsImage.getSuffix(
                inData["listH5FilePath"][0]
            )
            in_for_crystfel["image_directory"] = str(
                pathlib.Path(inData["listH5FilePath"][0]).parent
            )
            in_for_crystfel["maxchunksize"] = inData.get("batchSize", 300)

        elif "cbfFileInfo" in inData.keys():
            in_for_crystfel["maxchunksize"] = inData["cbfFileInfo"].get("batchSize", 10)
            in_for_crystfel["listofImages"] = inData["cbfFileInfo"].get(
                "listofImages", []
            )
            in_for_crystfel["image_directory"] = inData["cbfFileInfo"]["directory"]
            in_for_crystfel["prefix"] = inData["cbfFileInfo"]["template"].strip(
                "####.cbf"
            )
            in_for_crystfel["suffix"] = UtilsImage.getSuffix(
                inData["cbfFileInfo"]["template"]
            )
            if len(in_for_crystfel["listofImages"]) == 0:
                in_for_crystfel["ImageRange"] = (
                    inData["cbfFileInfo"]["startNo"],
                    inData["cbfFileInfo"]["endNo"],
                )
                FirstImage = os.path.join(
                    inData["cbfFileInfo"]["directory"],
                    inData["cbfFileInfo"]["template"].replace("####", "0001"),
                )
            else:
                FirstImage = in_for_crystfel["listofImages"][0]

            Image = Im(FirstImage)
            in_for_crystfel["detectorType"] = (
                Image.imobject.headers["detector_name"][0]
                + Image.imobject.headers["detector_name"][1]
            )
        else:
            logger.error("input json must have either listH5FilePath or cbfFileInfo")

        if doCBFtoH5:
            cxi_all = list(self.getWorkingDirectory().glob("dozor*cxi"))
            current = len(cxi_all) - 1
            in_for_crystfel["image_directory"] = self.getWorkingDirectory()
            in_for_crystfel["prefix"] = "dozor_%d." % current
            in_for_crystfel["suffix"] = "cxi"
            in_for_crystfel["peak_search"] = "cxi"
            in_for_crystfel["peak_info"] = "/data/peakinfo"
            in_for_crystfel["maxchunksize"] = 10

        in_for_crystfel["doSubmit"] = inData.get("doSubmit", True)

        crysttask = run_crystfel.AutoCrystFEL(in_for_crystfel)
        outstream = None
        results = dict()
        try:
            jsonschema.validate(
                instance=crysttask.jshandle, schema=crysttask.getInDataSchema()
            )
            crysttask.run_indexing()
            crysttask.combine_streams(doSubmit=doSubmit)

            outstream = crysttask.getOutputDirectory() / "alltogether.stream"
            results["QualityMetrics"] = crysttask.report_stats(str(outstream))
            print(results)
            crysttask.writeOutputData(results)
            '''
            crysttask.datafinder()
            crysttask.makeOutputDirectory()
            kk = {}
            if crysttask.jshandle['suffix'] == 'cxi':
                kk['cxi'] = """dim0 = %\ndim1 = ss\ndim2 = fs\ndata = /data/data\n"""
                geomfile = crysttask.make_geometry_file(**kk)
            else:
                geomfile = crysttask.make_geometry_file(**kk)
            print(geomfile)
            crysttask.make_list_events(str(geomfile))
            infile = str(crysttask.getOutputDirectory() / 'input.lst')
            outname = datetime.now().strftime('%H-%M-%S.stream')
            outstream = str(crysttask.getOutputDirectory() / outname)

            ofh = open(infile, 'w')
            for fname in crysttask.filelist:
                ofh.write(fname)
                ofh.write('\n')
            ofh.close()
            crystfel_cmd = crysttask.indexamajig_cmd(infile, outstream, geomfile)
            self.runCommandLine(crystfel_cmd, doSubmit=inData.get('doSubmit', True))
            '''
        except Exception as err:
            self.setFailure()
            logger.error(err)

        return outstream, results


class CrystFEL2ISPyB(AbstractTask):
    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "first_data_path": {"type": "string"},
                "dataCollectionId": {"type": "integer"},
                "streamfile": {"type": "string"},
                "centering": {"type": "string"},
                "num_indexed_frames": {"type": "integer"},
                "lattice": {"type": "string"},
                "unique_axis": {"type": "string"},
                "unit_cell": {
                    "type": "array",
                    "items": {"type": "number"},
                },
                "point_group": {"type": "string"},
                "space_group": {"type": "string"},
                "resolution_limit": {"type": "number"},
                "average_num_spots": {"type": "number"},
            },
        }

    def getOutDataSchema(self):
        return {"type": "object", "properties": {"status": {"type": "string"}}}

    def run(self, inData):
        qm = inData["QualityMetrics"]
        dataCollection_id = inData["dataCollectionId"]
        # Create ssx_cells.json
        ssx_cells = {"unit_cells": qm["unit_cell_array"]}
        working_dir = self.getWorkingDirectory()
        ssx_cells_path = working_dir / "ssx_cells.json"
        with open(ssx_cells_path, "w") as f:
            f.write(json.dumps(ssx_cells, indent=4))
        # Create ssx_stats.json
        ssx_stats = {
            "nbHits": qm["number_hits"],
            "nbIndexed": qm["num_indexed_frames"],
            "laticeType": qm["lattice"],
            "estimatedResolution": qm["resolution_limit"],
        }
        ssx_stats_path = working_dir / "ssx_stats.json"
        with open(ssx_stats_path, "w") as f:
            f.write(json.dumps(ssx_stats, indent=4))
        # Create pyarch directory
        first_data_path = pathlib.Path(inData["first_data_path"])
        pyarch_dir = UtilsPath.createPyarchFilePath(first_data_path.parent / first_data_path.stem)
        os.makedirs(pyarch_dir, mode=0o755, exist_ok=True)
        # Copy files to pyarch
        UtilsPath.systemCopyFile(ssx_cells_path, pyarch_dir)
        ssx_cells_pyarch_path = pyarch_dir / "ssx_cells.json"
        UtilsPath.systemCopyFile(ssx_stats_path, pyarch_dir)
        ssx_stats_pyarch_path = pyarch_dir / "ssx_stats.json"
        # Upload info to ISPyB
        #
        # 1. Create AutoProcProgram entry
        auto_proc_program_id = UtilsIspyb.storeOrUpdateAutoProcProgram(
            programs="CrystFEL",
            commandline="indexamajig",
            status="SUCCESS"
        )
        # 2. Create "dummy" AutoProcIntegration entry for linking with data collection
        auto_proc_integration_id = UtilsIspyb.storeOrUpdateAutoProcIntegration(
            auto_proc_program_id=auto_proc_program_id,
            dataCollection_id=dataCollection_id
        )
        # 3. Upload ssx_cells and ssx_stats as program attachments
        auto_proc_program_attachment_id_cells = UtilsIspyb.storeOrUpdateAutoProcProgramAttachment(
            auto_proc_program_id=auto_proc_program_id,
            file_type="Result",
            file_path=ssx_cells_pyarch_path
        )
        auto_proc_program_attachment_id_stats = UtilsIspyb.storeOrUpdateAutoProcProgramAttachment(
            auto_proc_program_id=auto_proc_program_id,
            file_type="Result",
            file_path=ssx_stats_pyarch_path
        )
        out_data = {
            "status": "ok",
            "pyarch_dir": pyarch_dir,
            "dataCollectionId": dataCollection_id,
            "auto_proc_program_id": auto_proc_program_id,
            "auto_proc_integration_id": auto_proc_integration_id,
            "auto_proc_program_attachment_id_cells": auto_proc_program_attachment_id_cells,
            "auto_proc_program_attachment_id_stats": auto_proc_program_attachment_id_stats
        }
        return out_data


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s %(name)-12s %(levelname)-8s%(message)s",
        datefmt="%y-%m-%d %H:%M",
        filename="autocryst.log",
        filemode="a+",
    )
    fh = open(sys.argv[1], "r")
    inData = json.load(fh)
    fh.close()
    crystfel = ExeCrystFEL(inData)

    crystfel.executeRun()
    print(crystfel.outData)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

import os
import json
import pathlib
import billiard
import traceback
import jsonschema
import subprocess

from edna2.utils import UtilsPath
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


class EDNA2Process(billiard.Process):
    """
    See https://stackoverflow.com/a/33599967.
    """

    def __init__(self, *args, **kwargs):
        billiard.Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = billiard.Pipe()
        self._exception = None

    def run(self):
        try:
            billiard.Process.run(self)
            self._cconn.send(None)
        except BaseException as e:
            tb = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


class AbstractTask():  # noqa R0904
    """
    Parent task to all EDNA2 tasks.
    """

    def __init__(self, inData, workingDirectorySuffix=None):
        self._dictInOut = billiard.Manager().dict()
        self._dictInOut["inData"] = json.dumps(inData, default=str)
        self._dictInOut["outData"] = json.dumps({})
        self._dictInOut["isFailure"] = False
        self._process = EDNA2Process(target=self.executeRun, args=())
        self._workingDirectorySuffix = workingDirectorySuffix
        self._workingDirectory = None
        self._logFileName = None
        self._errorLogFileName = None
        self._schemaPath = pathlib.Path(__file__).parents[1] / "schema"
        self._persistInOutData = True
        self._oldDir = os.getcwd()

    def getSchemaUrl(self, schemaName):
        return "file://" + str(self._schemaPath / schemaName)

    def executeRun(self):
        inData = self.getInData()
        hasValidInDataSchema = False
        hasValidOutDataSchema = False
        if self.getInDataSchema() is not None:
            instance = inData
            schema = self.getInDataSchema()
            try:
                jsonschema.validate(instance=instance, schema=schema)
                hasValidInDataSchema = True
            except Exception as e:
                logger.exception(e)
        else:
            hasValidInDataSchema = True
        if hasValidInDataSchema:
            self._workingDirectory = UtilsPath.getWorkingDirectory(
                self, inData, workingDirectorySuffix=self._workingDirectorySuffix
            )
            self.writeInputData(inData)
            self._oldDir = os.getcwd()
            os.chdir(str(self._workingDirectory))
            outData = self.run(inData)
            os.chdir(self._oldDir)
        else:
            raise RuntimeError("Schema validation error for inData")
        if self.getOutDataSchema() is not None:
            instance = outData
            schema = self.getOutDataSchema()
            try:
                jsonschema.validate(instance=instance, schema=schema)
                hasValidOutDataSchema = True
            except Exception as e:
                logger.exception(e)
        else:
            hasValidOutDataSchema = True
        if hasValidOutDataSchema:
            self.writeOutputData(outData)
        else:
            raise RuntimeError("Schema validation error for outData")
        if not os.listdir(str(self._workingDirectory)):
            os.rmdir(str(self._workingDirectory))

    def getInData(self):
        return json.loads(self._dictInOut["inData"])

    def setInData(self, inData):
        self._dictInOut["inData"] = json.dumps(inData, default=str)

    inData = property(getInData, setInData)

    def getOutData(self):
        return json.loads(self._dictInOut["outData"])

    def setOutData(self, outData):
        self._dictInOut["outData"] = json.dumps(outData, default=str)

    outData = property(getOutData, setOutData)

    def writeInputData(self, inData):
        # Write input data
        if self._persistInOutData and self._workingDirectory is not None:
            jsonName = "inData" + self.__class__.__name__ + ".json"
            with open(str(self._workingDirectory / jsonName), "w") as f:
                f.write(json.dumps(inData, default=str, indent=4))

    def writeOutputData(self, outData):
        self.setOutData(outData)
        if self._persistInOutData and self._workingDirectory is not None:
            jsonName = "outData" + self.__class__.__name__ + ".json"
            with open(str(self._workingDirectory / jsonName), "w") as f:
                f.write(json.dumps(outData, default=str, indent=4))

    def getLogPath(self):
        if self._logFileName is None:
            self._logFileName = self.__class__.__name__ + ".log.txt"
        logPath = self._workingDirectory / self._logFileName
        return logPath

    def setLogFileName(self, logFileName):
        self._logFileName = logFileName

    def getLogFileName(self):
        return self._logFileName

    def getErrorLogPath(self):
        if self._errorLogFileName is None:
            self._errorLogFileName = self.__class__.__name__ + ".error.txt"
        errorLogPath = self._workingDirectory / self._errorLogFileName
        return errorLogPath

    def setErrorLogFileName(self, errorLogFileName):
        self._errorLogFileName = errorLogFileName

    def getErrorLogFileName(self):
        return self._errorLogFileName

    def getLog(self):
        with open(str(self.getLogPath())) as f:
            log = f.read()
        return log

    def getErrorLog(self):
        with open(str(self.getErrorLogPath())) as f:
            errorLog = f.read()
        return errorLog

    def submitCommandLine(self, commandLine, jobName, partition, ignoreErrors, noCores=None):
        workingDir = str(self._workingDirectory)
        if workingDir.startswith("/mntdirect/_users"):
            workingDir = workingDir.replace("/mntdirect/_users", "/home/esrf")
        nodes = 1
        time = "1:00:00"
        mem = 16000  # 16 Gb memory by default
        script = "#!/bin/bash\n"
        script += '#SBATCH --job-name="{0}"\n'.format(jobName)
        if partition is None:
            partition = "mx"
        else:
            partition = "{0}".format(partition)
        script += "#SBATCH --partition={0}\n".format(partition)
        script += "#SBATCH --mem={0}\n".format(mem)
        script += "#SBATCH --nodes={0}\n".format(nodes)
        if noCores is None:
            script += "#SBATCH --exclusive\n"
        else:
            script += "#SBATCH --cpus-per-task={0}\n".format(noCores)
        script += "#SBATCH --time={0}\n".format(time)
        script += "#SBATCH --chdir={0}\n".format(workingDir)
        script += "#SBATCH --output=stdout.txt\n"
        script += "#SBATCH --error=stderr.txt\n"
        script += commandLine + "\n"
        shellFile = self._workingDirectory / (jobName + "_slurm.sh")
        with open(str(shellFile), "w") as f:
            f.write(script)
            f.close()
        shellFile.chmod(0o755)
        slurmCommandLine = "sbatch --wait {0}".format(shellFile)
        pipes = subprocess.Popen(
            slurmCommandLine,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            cwd=str(self._workingDirectory),
        )
        stdout, stderr = pipes.communicate()
        slurmLogPath = self._workingDirectory / (jobName + "_slurm.log")
        slurmErrorLogPath = self._workingDirectory / (jobName + "_slurm.error.log")
        if len(stdout) > 0:
            log = str(stdout, "utf-8")
            with open(str(slurmLogPath), "w") as f:
                f.write(log)
        if len(stderr) > 0:
            if not ignoreErrors:
                logger.warning(
                    "Error messages from command {0}".format(commandLine.split(" ")[0])
                )
            with open(str(slurmErrorLogPath), "w") as f:
                f.write(str(stderr, "utf-8"))
        if pipes.returncode != 0:
            # Error!
            warningMessage = "{0}, code {1}".format(stderr, pipes.returncode)
            logger.warning(warningMessage)
            # raise RuntimeError(errorMessage)

    def runCommandLine(
        self,
        commandLine,
        logPath=None,
        listCommand=None,
        ignoreErrors=False,
        doSubmit=False,
        partition=None,
        noCores=None
    ):
        if logPath is None:
            logPath = self.getLogPath()
        jobName = self.__class__.__name__
        logFileName = os.path.basename(logPath)
        errorLogPath = self.getErrorLogPath()
        errorLogFileName = os.path.basename(errorLogPath)
        commandLine += " 1>{0} 2>{1}".format(logFileName, errorLogFileName)
        if listCommand is not None:
            commandLine += " << EOF-EDNA2\n"
            for command in listCommand:
                commandLine += command + "\n"
            commandLine += "EOF-EDNA2"
        commandLogFileName = jobName + ".commandLine.txt"
        commandLinePath = self._workingDirectory / commandLogFileName
        with open(str(commandLinePath), "w") as f:
            f.write(commandLine)
        if doSubmit:
            self.submitCommandLine(commandLine, jobName, partition, ignoreErrors, noCores)
        else:
            pipes = subprocess.Popen(
                commandLine,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                close_fds=True,
                cwd=str(self._workingDirectory),
            )
            stdout, stderr = pipes.communicate()
            if len(stdout) > 0:
                log = str(stdout, "utf-8")
                with open(str(logPath), "w") as f:
                    f.write(log)
            if len(stderr) > 0:
                if not ignoreErrors:
                    logger.warning(
                        "Error messages from command {0}".format(
                            commandLine.split(" ")[0]
                        )
                    )
                errorLogPath = self._workingDirectory / errorLogFileName
                with open(str(errorLogPath), "w") as f:
                    f.write(str(stderr, "utf-8"))
            if pipes.returncode != 0:
                # Error!
                errorMessage = "{0}, code {1}".format(stderr, pipes.returncode)
                raise RuntimeError(errorMessage)

    def onError(self):
        pass

    def start(self):
        self._process.start()

    def join(self):
        self._process.join()
        if self._process.exception:
            error, trace = self._process.exception
            logger.error(error)
            logger.error(trace)
            self._dictInOut["isFailure"] = True
            self.onError()

    def execute(self):
        self.start()
        self.join()

    def setFailure(self):
        self._dictInOut["isFailure"] = True

    def isFailure(self):
        return self._dictInOut["isFailure"]

    def isSuccess(self):
        return not self.isFailure()

    def getWorkingDirectory(self):
        return self._workingDirectory

    def setWorkingDirectory(self, inData):
        self._workingDirectory = UtilsPath.getWorkingDirectory(self, inData)

    def getInDataSchema(self):
        return None

    def getOutDataSchema(self):
        return None

    def setPersistInOutData(self, value):
        self._persistInOutData = value




#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "20/04/2020"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx

# mxPluginExec/plugins/EDPluginGroupXDS-v1.0/plugins/EDPluginXDSv1_0.py
# mxPluginExec/plugins/EDPluginGroupXDS-v1.0/plugins/EDPluginXDSIndexingv1_0.py

import os
import math
import shutil
import numpy as np

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging
from edna2.utils import UtilsDetector
from edna2.utils import UtilsSymmetry


logger = UtilsLogging.getLogger()


R2D = 180 / math.pi


class XDSTask(AbstractTask):
    """
    Common base class for all XDS tasks
    """

    def run(self, inData):
        commandLine = "xds_par"
        listXDS_INP = self.generateXDS_INP(inData)
        self.writeXDS_INP(listXDS_INP, self.getWorkingDirectory())
        self.setLogFileName("xds.log")
        self.runCommandLine(commandLine, listCommand=[])
        # Work in progress!
        outData = self.parseXDSOutput(self.getWorkingDirectory())
        return outData

    @staticmethod
    def generateImageLinks(in_data, working_directory=None):
        first_sub_wedge = in_data["subWedge"][0]
        first_image_path = first_sub_wedge["image"][0]["path"]
        prefix = UtilsImage.getPrefix(first_image_path)
        suffix = UtilsImage.getSuffix(first_image_path)
        if suffix == "h5":
            lowest_xds_image_number = 1
            highest_xds_image_number = 1
            h5MasterFilePath, h5DataFilePath, h5FileNumber = UtilsImage.getH5FilePath(
                first_image_path, hasOverlap=True, isFastMesh=False
            )
            h5MasterFile = os.path.basename((str(h5MasterFilePath)))
            h5DataFile = os.path.basename((str(h5DataFilePath)))
            list_image_link = [
                [str(h5MasterFilePath), h5MasterFile],
                [str(h5DataFilePath), h5DataFile],
            ]
            if working_directory is not None:
                os.symlink(str(h5MasterFilePath), h5MasterFile)
                os.symlink(str(h5DataFilePath), h5DataFile)
            template = h5MasterFile.replace("master", "??????")
            lowest_xds_image_number = None
            highest_xds_image_number = None
            for subwedge in in_data["subWedge"]:
                image_list = subwedge["image"]
                for image_dict in image_list:
                    image_path = image_dict["path"]
                    image_number = UtilsImage.getImageNumber(image_path)
                    if lowest_xds_image_number is None or lowest_xds_image_number > image_number:
                        lowest_xds_image_number = image_number
                    if highest_xds_image_number is None or highest_xds_image_number < image_number:
                        highest_xds_image_number = image_number
        else:
            template = "%s_xdslink_?????.%s" % (prefix, suffix)
            xds_lowest_image_number_global = 1
            # First we have to find the smallest goniostat rotation axis start:
            oscillation_start_min = 0

            # Loop through the list of sub wedges
            list_of_list = []
            lowest_xds_image_number = None
            highest_xds_image_number = None
            list_spot_range = []
            for sub_wedge in in_data["subWedge"]:
                list_image_link = []
                image_list = sub_wedge["image"]
                goniostat = sub_wedge["experimentalCondition"]["goniostat"]
                oscillation_start = goniostat["rotationAxisStart"]
                oscillation_range = goniostat["oscillationWidth"]

                # First find the lowest and highest image numbers
                lowest_image_number = None
                for image in image_list:
                    image_number = image["number"]
                    if lowest_image_number is None or image_number < lowest_image_number:
                        lowest_image_number = image_number

                # Loop through the list of images
                spot_range_min = None
                spot_range_max = None
                for image in image_list:
                    image_number = image["number"]
                    image_oscillation_start = (
                        oscillation_start
                        + (image_number - lowest_image_number) * oscillation_range
                    )
                    # if xdsLowestImageNumberGlobal is None:
                    #     xdsLowestImageNumberGlobal = 1 + int((imageOscillationStart - oscillationStartMin) / oscillationRange)
                    xds_image_number = xds_lowest_image_number_global + int(
                        (image_oscillation_start - oscillation_start_min) / oscillation_range + 0.5
                    )
                    print(
                        xds_image_number,
                        image_oscillation_start,
                        oscillation_start_min,
                        oscillation_range,
                    )
                    source_path = image["path"]
                    target = "%s_xdslink_%05d.%s" % (prefix, xds_image_number, suffix)
                    print([source_path, target])
                    list_image_link.append([source_path, target])
                    if working_directory is not None and not os.path.exists(target):
                        os.symlink(source_path, target)
                    if (
                        lowest_xds_image_number is None
                        or lowest_xds_image_number > xds_image_number
                    ):
                        lowest_xds_image_number = xds_image_number
                    if (
                        highest_xds_image_number is None
                        or highest_xds_image_number < xds_image_number
                    ):
                        highest_xds_image_number = xds_image_number
                    if spot_range_min is None or spot_range_min > xds_image_number:
                        spot_range_min = xds_image_number
                    if spot_range_max is None or spot_range_max < xds_image_number:
                        spot_range_max = xds_image_number
                list_spot_range.append([spot_range_min, spot_range_max])
                list_of_list.append(list_image_link)
        previous_exclude_data_range_max = 1
        list_exclude_data_range = []
        for spot_range_min, spot_range_max in list_spot_range:
            if spot_range_min > previous_exclude_data_range_max + 1:
                list_exclude_data_range.append([previous_exclude_data_range_max, spot_range_min - 1])
            previous_exclude_data_range_max = spot_range_max + 1
        dictImageLinks = {
            "imageLink": list_of_list,
            "spotRange": list_spot_range,
            "dataRange": [lowest_xds_image_number, highest_xds_image_number],
            "excludeDataRange": list_exclude_data_range,
            "template": template,
        }
        return dictImageLinks

    @staticmethod
    def generateXDS_INP(inData):
        """
        This method creates a list of XDS.INP commands
        """
        # Take the first sub webge in input as reference
        firstSubwedge = inData["subWedge"][0]
        listImage = firstSubwedge["image"]
        image = listImage[0]
        experimentalCondition = firstSubwedge["experimentalCondition"]
        detector = experimentalCondition["detector"]
        dictXDSDetector = XDSTask.getXDSDetector(detector)
        beam = experimentalCondition["beam"]
        goniostat = experimentalCondition["goniostat"]
        distance = round(detector["distance"], 3)
        wavelength = round(beam["wavelength"], 3)
        oscRange = goniostat["oscillationWidth"]
        startAngle = goniostat["rotationAxisStart"] - int(
            goniostat["rotationAxisStart"]
        )
        listXDS_INP = [
            "OVERLOAD=10048500",
            "DIRECTION_OF_DETECTOR_X-AXIS={0}".format(
                UtilsConfig.get("XDSTask", "DIRECTION_OF_DETECTOR_X-AXIS")
            ),
            "DIRECTION_OF_DETECTOR_Y-AXIS={0}".format(
                UtilsConfig.get("XDSTask", "DIRECTION_OF_DETECTOR_Y-AXIS")
            ),
            "ROTATION_AXIS={0}".format(UtilsConfig.get("XDSTask", "ROTATION_AXIS")),
            "INCIDENT_BEAM_DIRECTION={0}".format(
                UtilsConfig.get("XDSTask", "INCIDENT_BEAM_DIRECTION")
            ),
            "NX={0} NY={1} QX={2} QY={2}".format(
                dictXDSDetector["nx"], dictXDSDetector["ny"], dictXDSDetector["pixel"]
            ),
            "ORGX={0} ORGY={1}".format(
                dictXDSDetector["orgX"], dictXDSDetector["orgY"]
            ),
            "DETECTOR={0}  MINIMUM_VALID_PIXEL_VALUE={1}  OVERLOAD={2}".format(
                dictXDSDetector["name"],
                dictXDSDetector["minimumValidPixelValue"],
                dictXDSDetector["overload"],
            ),
            "SENSOR_THICKNESS={0}".format(dictXDSDetector["sensorThickness"]),
            "TRUSTED_REGION={0} {1}".format(
                dictXDSDetector["trustedRegion"][0], dictXDSDetector["trustedRegion"][1]
            ),
        ]
        # for trustedRegion in dictXDSDetector["untrustedRectangle"]:
        #     listXDS_INP.append('UNTRUSTED_RECTANGLE={0} {1} {2} {3}'.format(
        #         trustedRegion[0], trustedRegion[1],
        #         trustedRegion[2],trustedRegion[3]
        #     ))
        listXDS_INP += [
            "DETECTOR_DISTANCE={0}".format(distance),
            "X-RAY_WAVELENGTH={0}".format(wavelength),
            "OSCILLATION_RANGE={0}".format(oscRange),
            "STARTING_ANGLE={0}".format(startAngle),
            "INDEX_QUALITY= 0.25",
        ]
        if "spaceGroupNumber" in inData:
            spaceGroupNumber = inData["spaceGroupNumber"]
            cell = inData["cell"]
            unitCellConstants = "{a} {b} {c} {alpha} {beta} {gamma}".format(**cell)
            listXDS_INP += [
                "SPACE_GROUP_NUMBER={0}".format(spaceGroupNumber),
                "UNIT_CELL_CONSTANTS={0}".format(unitCellConstants),
            ]
        if (
            image["path"].endswith("h5")
            and UtilsConfig.get("XDSTask", "LIB") is not None
        ):
            listXDS_INP += ["LIB= {0}".format(UtilsConfig.get("XDSTask", "LIB"))]
        return listXDS_INP

    @staticmethod
    def createSPOT_XDS(listDozorSpotFile, oscRange):
        """
              implicit none
              integer nmax
              parameter(nmax=10000000)
              real*4 x(3),j
              integer n,i,k
              real*4 xa(nmax,3),ja(nmax)
              logical new
        c
              n=0
              do while(.true.)
                 read(*,*,err=1,end=1)x,j
                 new = .true.
                 do i = n,1,-1
                    if (abs(xa(i,3)-x(3)) .gt. 20.0 ) goto 3
                    do k = 1,2
                       if (abs(x(k)-xa(i,k)) .gt. 6.0) goto 2
                    enddo
                    new = .false.
                    xa(i,:)=(xa(i,:)*ja(i)+x*j)/(ja(i)+j)
                    ja(i)=ja(i)+j
          2         continue
                 enddo
          3       if (new) then
                    n=n+1
                    xa(n,:)=x
                    ja(n)=j
                 endif
              enddo
          1   continue
              do i=1,n
                 write(*,*)xa(i,:), ja(i)
              enddo
              end
        """
        listSpotXds = []
        n = 0
        for dozorSpotFile in listDozorSpotFile:
            # Read the file
            with open(str(dozorSpotFile)) as f:
                dozorLines = f.readlines()
            omega = float(dozorLines[2].split()[1]) % 360
            frame = int((omega - oscRange / 2) / oscRange) + 1
            print(omega, frame)
            for dozorLine in dozorLines[3:]:
                new = True
                listValues = dozorLine.split()
                n, xPos, yPos, intensity, sigma = list(map(float, listValues))
                # Subtracting 1 from X and Y: this is because for dozor the upper left pixel in the image is (1,1),
                # whereas for the rest of the world it is (0,0)
                xPos = xPos - 1
                yPos = yPos - 1
                index = 0
                for spotXds in listSpotXds:
                    frameOld = spotXds[2]
                    if abs(frameOld - frame) > 20:
                        break
                    xPosOld = spotXds[0]
                    yPosOld = spotXds[1]
                    intensityOld = spotXds[3]
                    if abs(xPosOld - xPos) <= 6 and abs(yPosOld - yPos) <= 6:
                        new = False
                        intensityNew = intensity + intensityOld
                        xPosNew = (
                            xPosOld * intensityOld + xPos * intensity
                        ) / intensityNew
                        yPosNew = (
                            yPosOld * intensityOld + yPos * intensity
                        ) / intensityNew
                        listSpotXds[index] = [xPosNew, yPosNew, frameOld, intensityNew]
                    index += 1

                if new:
                    spotXds = [xPos, yPos, frame, intensity]
                    listSpotXds.append(spotXds)

        strSpotXds = ""
        for spotXds in listSpotXds:
            strSpotXds += "{0:13.6f}{1:17.6f}{2:17.8f}{3:17.6f}    \n".format(*spotXds)
        return strSpotXds

    @staticmethod
    def writeSPOT_XDS(listDozorSpotFile, oscRange, workingDirectory):
        spotXds = XDSTask.createSPOT_XDS(listDozorSpotFile, oscRange)
        filePath = workingDirectory / "SPOT.XDS"
        with open(str(filePath), "w") as f:
            f.write(spotXds)

    def writeXDS_INP(self, listXDS_INP, workingDirectory):
        fileName = "XDS.INP"
        filePath = workingDirectory / fileName
        with open(str(filePath), "w") as f:
            for line in listXDS_INP:
                f.write(line + "\n")

    @staticmethod
    def getXDSDetector(dictDetector):
        untrustedRectangle = []
        dictXDSDetector = None
        detectorType = dictDetector["type"]
        nx = UtilsDetector.getNx(detectorType)
        ny = UtilsDetector.getNy(detectorType)
        pixel = UtilsDetector.getPixelsize(detectorType)
        orgX = round(dictDetector["beamPositionX"] / pixel, 3)
        orgY = round(dictDetector["beamPositionY"] / pixel, 3)
        if detectorType == "pilatus2m":
            untrustedRectangle = [
                [487, 495, 0, 1680],
                [981, 989, 0, 1680],
                [0, 1476, 195, 213],
                [0, 1476, 407, 425],
                [0, 1476, 619, 637],
                [0, 1476, 831, 849],
                [0, 1476, 1043, 1061],
                [0, 1476, 1255, 1273],
                [0, 1476, 1467, 1485],
            ]
            sensorThickness = 0.32
        elif detectorType == "pilatus6m":
            untrustedRectangle = [
                [487, 495, 0, 2528],
                [981, 989, 0, 2528],
                [1475, 1483, 0, 2528],
                [1969, 1977, 0, 2528],
                [0, 2464, 195, 213],
                [0, 2464, 407, 425],
                [0, 2464, 619, 637],
                [0, 2464, 831, 849],
                [0, 2464, 1043, 1061],
                [0, 2464, 1255, 1273],
                [0, 2464, 1467, 1485],
                [0, 2464, 1679, 1697],
                [0, 2464, 1891, 1909],
                [0, 2464, 2103, 2121],
                [0, 2464, 2315, 2333],
            ]
            sensorThickness = 0.32
        elif detectorType == "eiger4m":
            untrustedRectangle = [
                [1029, 1040, 0, 2167],
                [0, 2070, 512, 550],
                [0, 2070, 1063, 1103],
                [0, 2070, 1614, 1654],
            ]
            sensorThickness = 0.32
        elif detectorType == "eiger9m":
            untrustedRectangle = [
                [1029, 1040, 0, 3269],
                [2069, 2082, 0, 3269],
                [0, 3110, 513, 553],
                [0, 3110, 1064, 1104],
                [0, 3110, 1615, 1655],
                [0, 3110, 2166, 2206],
                [0, 3110, 2717, 2757],
            ]
        elif detectorType == "eiger16m":
            untrustedRectangle = [
                [0, 4149, 512, 549],
                [0, 4149, 1062, 1099],
                [0, 4149, 1612, 1649],
                [0, 4149, 2162, 2199],
                [0, 4149, 2712, 2749],
                [0, 4149, 3262, 3299],
                [0, 4149, 3812, 3849],
                [513, 514, 0, 4362],
                [1028, 1039, 0, 4362],
                [1553, 1554, 0, 4362],
                [2068, 2079, 0, 4362],
                [2593, 2594, 0, 4362],
                [3108, 3119, 0, 4362],
                [3633, 3634, 0, 4362],
            ]
            sensorThickness = 0.75
        else:
            raise RuntimeError("Unknown detector: {0}".format(detectorType))
        dictXDSDetector = {
            "name": "PILATUS",
            "nx": nx,
            "ny": ny,
            "orgX": orgX,
            "orgY": orgY,
            "pixel": pixel,
            "untrustedRectangle": untrustedRectangle,
            "trustedRegion": [0.0, 1.41],
            "trustedpixel": [7000, 30000],
            "minimumValidPixelValue": 0,
            "overload": 1048500,
            "sensorThickness": sensorThickness,
        }
        return dictXDSDetector


class XDSIndexing(XDSTask):
    def generateXDS_INP(self, inData):
        first_sub_wedge = inData["subWedge"][0]
        # listDozorSpotFile = inData["dozorSpotFile"]
        experimental_condition = first_sub_wedge["experimentalCondition"]
        goniostat = experimental_condition["goniostat"]
        oscRange = goniostat["oscillationWidth"]
        # XDSTask.writeSPOT_XDS(
        #     listDozorSpotFile,
        #     oscRange=oscRange,
        #     workingDirectory=self.getWorkingDirectory(),
        # )
        list_xds_inp = XDSTask.generateXDS_INP(inData)
        list_xds_inp.insert(0, "JOB= XYCORR INIT COLSPOT IDXREF")
        dict_image_links = self.generateImageLinks(inData, self.getWorkingDirectory())
        list_xds_inp.append(
            "NAME_TEMPLATE_OF_DATA_FRAMES= {0}".format(dict_image_links["template"])
        )
        list_spot_range = dict_image_links["spotRange"]
        for spot_range_min, spot_range_max in list_spot_range:
            list_xds_inp.append(
                "SPOT_RANGE= {0} {1}".format(spot_range_min, spot_range_max)
            )
        list_xds_inp.append(
            "DATA_RANGE= {0} {1}".format(
                dict_image_links["dataRange"][0], dict_image_links["dataRange"][1]
            )
        )
        list_spot_range = dict_image_links["excludeDataRange"]
        for exclude_range_min, exclude_range_max in list_spot_range:
            list_xds_inp.append(
                "EXCLUDE_DATA_RANGE= {0} {1}".format(exclude_range_min, exclude_range_max)
            )
        return list_xds_inp

    @staticmethod
    def parseXDSOutput(workingDirectory):
        idxref_path = workingDirectory / "IDXREF.LP"
        xparm_path = workingDirectory / "XPARM.XDS"
        spot_path = workingDirectory / "SPOT.XDS"
        out_data = {
            "idxref": XDSIndexing.readIdxrefLp(idxref_path),
            "xparm": XDSIndexing.parseXparm(xparm_path),
            "xparmXdsPath": xparm_path,
            "spotXdsPath": spot_path
        }
        return out_data

    @staticmethod
    def parseParameters(indexLine, listLines, resultXDSIndexing):
        if "MOSAICITY" in listLines[indexLine]:
            resultXDSIndexing["mosaicity"] = float(listLines[indexLine].split()[-1])
        elif "DETECTOR COORDINATES (PIXELS) OF DIRECT BEAM" in listLines[indexLine]:
            resultXDSIndexing["xBeam"] = float(listLines[indexLine].split()[-1])
            resultXDSIndexing["yBeam"] = float(listLines[indexLine].split()[-2])
        elif "CRYSTAL TO DETECTOR DISTANCE" in listLines[indexLine]:
            resultXDSIndexing["distance"] = float(listLines[indexLine].split()[-1])

    @staticmethod
    def parseLattice(indexLine, listLines, resultXDSIndexing):
        if listLines[indexLine].startswith(" * ") and not listLines[
            indexLine + 1
        ].startswith(" * "):
            listLine = listLines[indexLine].split()
            latticeCharacter = int(listLine[1])
            bravaisLattice = listLine[2]
            spaceGroup = UtilsSymmetry.getMinimumSymmetrySpaceGroupFromBravaisLattice(
                bravaisLattice
            )
            spaceGroupNumber = UtilsSymmetry.getITNumberFromSpaceGroupName(spaceGroup)
            qualityOfFit = float(listLine[3])
            resultXDSIndexing.update(
                {
                    "latticeCharacter": latticeCharacter,
                    "spaceGroupNumber": spaceGroupNumber,
                    "qualityOfFit": qualityOfFit,
                    "a": float(listLine[4]),
                    "b": float(listLine[5]),
                    "c": float(listLine[6]),
                    "alpha": float(listLine[7]),
                    "beta": float(listLine[8]),
                    "gamma": float(listLine[9]),
                }
            )

    @staticmethod
    def readIdxrefLp(pathToIdxrefLp, resultXDSIndexing=None):
        if resultXDSIndexing is None:
            resultXDSIndexing = {}
        if pathToIdxrefLp.exists():
            with open(str(pathToIdxrefLp)) as f:
                listLines = f.readlines()
            indexLine = 0
            doParseParameters = False
            doParseLattice = False
            while indexLine < len(listLines):
                if (
                    "DIFFRACTION PARAMETERS USED AT START OF INTEGRATION"
                    in listLines[indexLine]
                ):
                    doParseParameters = True
                    doParseLattice = False
                elif (
                    "DETERMINATION OF LATTICE CHARACTER AND BRAVAIS LATTICE"
                    in listLines[indexLine]
                ):
                    doParseParameters = False
                    doParseLattice = True
                if doParseParameters:
                    XDSIndexing.parseParameters(indexLine, listLines, resultXDSIndexing)
                elif doParseLattice:
                    XDSIndexing.parseLattice(indexLine, listLines, resultXDSIndexing)
                indexLine += 1
        return resultXDSIndexing

    @staticmethod
    def volum(cell):
        """
        Calculate the cell volum from either:
         - the 6 standard cell parameters (a, b, c, alpha, beta, gamma)
         - or the 3 vectors A, B, C
        Inspired from XOconv written by Pierre Legrand:
        https://github.com/legrandp/xdsme/blob/67001a75f3c363bfe19b8bd7cae999f4fb9ad49d/XOconv/XOconv.py#L758"""
        if len(cell) == 6 and isinstance(cell[0], float):
            # expect a, b, c, alpha, beta, gamma (angles in degree).
            ca, cb, cg = map(XDSIndexing.cosd, cell[3:6])
            return (
                cell[0]
                * cell[1]
                * cell[2]
                * (1 - ca ** 2 - cb ** 2 - cg ** 2 + 2 * ca * cb * cg) ** 0.5
            )
        elif len(cell) == 3 and isinstance(cell[0], np.array):
            # expect vectors of the 3 cell parameters
            A, B, C = cell
            return A * B.cross(C)
        else:
            return "Can't parse input arguments."

    @staticmethod
    def cosd(a):
        return math.cos(a / R2D)

    @staticmethod
    def sind(a):
        return math.sin(a / R2D)

    @staticmethod
    def reciprocal(cell):
        """
        Calculate the 6 reciprocal cell parameters: a*, b*, c*, alpha*, beta*...
        Inspired from XOconv written by Pierre Legrand:
        https://github.com/legrandp/xdsme/blob/67001a75f3c363bfe19b8bd7cae999f4fb9ad49d/XOconv/XOconv.py#L776
        """
        sa, sb, sg = map(XDSIndexing.sind, cell[3:6])
        ca, cb, cg = map(XDSIndexing.cosd, cell[3:6])
        v = XDSIndexing.volum(cell)
        rc = (
            cell[1] * cell[2] * sa / v,
            cell[2] * cell[0] * sb / v,
            cell[0] * cell[1] * sg / v,
            math.acos((cb * cg - ca) / (sb * sg)) * R2D,
            math.acos((ca * cg - cb) / (sa * sg)) * R2D,
            math.acos((ca * cb - cg) / (sa * sb)) * R2D,
        )
        return rc

    @staticmethod
    def BusingLevy(rcell):
        """
        Inspired from XOconv written by Pierre Legrand:
        https://github.com/legrandp/xdsme/blob/67001a75f3c363bfe19b8bd7cae999f4fb9ad49d/XOconv/XOconv.py#L816
        """
        ex = np.array([1, 0, 0])
        ey = np.array([0, 1, 0])
        cosr = list(map(XDSIndexing.cosd, rcell[3:6]))
        sinr = list(map(XDSIndexing.sind, rcell[3:6]))
        Vr = XDSIndexing.volum(rcell)
        BX = ex * rcell[0]
        BY = rcell[1] * (ex * cosr[2] + ey * sinr[2])
        c = rcell[0] * rcell[1] * sinr[2] / Vr
        cosAlpha = (cosr[1] * cosr[2] - cosr[0]) / (sinr[1] * sinr[2])
        BZ = np.array([rcell[2] * cosr[1], -1 * rcell[2] * sinr[1] * cosAlpha, 1 / c])
        return np.array([BX, BY, BZ]).transpose()

    @staticmethod
    def parseXparm(pathToXparmXds):
        """
        Inspired from parse_xparm written by Pierre Legrand:
        https://github.com/legrandp/xdsme/blob/67001a75f3c363bfe19b8bd7cae999f4fb9ad49d/XOconv/XOconv.py#L372
        """
        if pathToXparmXds.exists():
            with open(str(pathToXparmXds)) as f:
                xparm = f.readlines()
            xparamDict = {
                "rot": list(map(float, xparm[1].split()[3:])),
                "beam": list(map(float, xparm[2].split()[1:])),
                "distance": float(xparm[8].split()[2]),
                "originXDS": list(map(float, xparm[8].split()[:2])),
                "A": list(map(float, xparm[4].split())),
                "B": list(map(float, xparm[5].split())),
                "C": list(map(float, xparm[6].split())),
                "cell": list(map(float, xparm[3].split()[1:])),
                "pixel_size": list(map(float, xparm[7].split()[3:])),
                "pixel_numb": list(map(float, xparm[7].split()[1:])),
                "symmetry": int(xparm[3].split()[0]),
                "num_init": list(map(float, xparm[1].split()[:3]))[0],
                "phi_init": list(map(float, xparm[1].split()[:3]))[1],
                "delta_phi": list(map(float, xparm[1].split()[:3]))[2],
                "detector_X": list(map(float, xparm[9].split())),
                "detector_Y": list(map(float, xparm[10].split())),
            }
        else:
            xparamDict = {}
        return xparamDict


class XDSGenerateBackground(XDSTask):
    def generateXDS_INP(self, inData):
        listXDS_INP = XDSTask.generateXDS_INP(inData)
        listXDS_INP.insert(0, "JOB= XYCORR INIT COLSPOT")
        dictImageLinks = self.generateImageLinks(inData, self.getWorkingDirectory())
        listXDS_INP.append(
            "NAME_TEMPLATE_OF_DATA_FRAMES= {0}".format(dictImageLinks["template"])
        )
        listXDS_INP.append(
            "DATA_RANGE= {0} {1}".format(
                dictImageLinks["dataRange"][0], dictImageLinks["dataRange"][1]
            )
        )
        return listXDS_INP

    @staticmethod
    def parseXDSOutput(workingDirectory):
        if (workingDirectory / "BKGINIT.cbf").exists():
            outData = {
                "gainCbf": str(workingDirectory / "GAIN.cbf"),
                "blankCbf": str(workingDirectory / "BLANK.cbf"),
                "bkginitCbf": str(workingDirectory / "BKGINIT.cbf"),
                "xCorrectionsCbf": str(workingDirectory / "X-CORRECTIONS.cbf"),
                "yCorrectionsCbf": str(workingDirectory / "Y-CORRECTIONS.cbf"),
            }
        else:
            outData = {}
        return outData


class XDSIntegration(XDSTask):
    def generateXDS_INP(self, inData):
        # Copy XPARM.XDS, GAIN.CBF file
        shutil.copy(inData["xparmXds"], self.getWorkingDirectory())
        shutil.copy(inData["gainCbf"], self.getWorkingDirectory())
        shutil.copy(inData["xCorrectionsCbf"], self.getWorkingDirectory())
        shutil.copy(inData["yCorrectionsCbf"], self.getWorkingDirectory())
        shutil.copy(inData["blankCbf"], self.getWorkingDirectory())
        shutil.copy(inData["bkginitCbf"], self.getWorkingDirectory())
        listXDS_INP = XDSTask.generateXDS_INP(inData)
        listXDS_INP.insert(0, "JOB= DEFPIX INTEGRATE CORRECT")
        dictImageLinks = self.generateImageLinks(inData, self.getWorkingDirectory())
        listXDS_INP.append(
            "NAME_TEMPLATE_OF_DATA_FRAMES= {0}".format(dictImageLinks["template"])
        )
        listXDS_INP.append(
            "DATA_RANGE= {0} {1}".format(
                dictImageLinks["dataRange"][0], dictImageLinks["dataRange"][1]
            )
        )
        return listXDS_INP

    @staticmethod
    def parseXDSOutput(workingDirectory):
        outData = {}
        if (workingDirectory / "XDS_ASCII.HKL").exists():
            outData = {
                "xdsAsciiHkl": str(workingDirectory / "XDS_ASCII.HKL"),
                "correctLp": str(workingDirectory / "CORRECT.LP"),
                "bkgpixCbf": str(workingDirectory / "BKGPIX.cbf"),
            }
        return outData


class XDSIndexAndIntegration(XDSTask):
    def generateXDS_INP(self, inData):
        listXDS_INP = XDSTask.generateXDS_INP(inData)
        listXDS_INP.insert(0, "JOB= XYCORR INIT IDXREF COLSPOT DEFPIX INTEGRATE CORRECT")
        dictImageLinks = self.generateImageLinks(inData, self.getWorkingDirectory())
        listXDS_INP.append(
            "NAME_TEMPLATE_OF_DATA_FRAMES= {0}".format(dictImageLinks["template"])
        )
        no_background_images = min(
            (dictImageLinks["dataRange"][1] - dictImageLinks["dataRange"][0]),
            4
        )
        listXDS_INP.append("BACKGROUND_RANGE= {0} {1}".format(
            dictImageLinks["dataRange"][0],
            dictImageLinks["dataRange"][0]+no_background_images-1
        ))
        listXDS_INP.append(
            "SPOT_RANGE= {0} {1}".format(
                dictImageLinks["dataRange"][0], dictImageLinks["dataRange"][1]
            )
        )
        listXDS_INP.append(
            "DATA_RANGE= {0} {1}".format(
                dictImageLinks["dataRange"][0], dictImageLinks["dataRange"][1]
            )
        )
        return listXDS_INP

    @staticmethod
    def parseXDSOutput(workingDirectory):
        outData = {}
        if (workingDirectory / "XDS_ASCII.HKL").exists():
            outData = {
                "xdsAsciiHkl": str(workingDirectory / "XDS_ASCII.HKL"),
                "correctLp": str(workingDirectory / "CORRECT.LP"),
                "bkgpixCbf": str(workingDirectory / "BKGPIX.cbf"),
            }
        return outData



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxv1/src/EDHandlerESRFPyarchv1_0.py

import os
import subprocess
import time
import pathlib
import tempfile

from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()

DEFAULT_TIMEOUT = 120  # s


def getWorkingDirectory(task, inData, workingDirectorySuffix=None):
    parentDirectory = inData.get("workingDirectory", None)
    if parentDirectory is None:
        parentDirectory = os.getcwd()
    parentDirectory = pathlib.Path(parentDirectory)
    if not parentDirectory.exists():
        parentDirectory.mkdir(mode=0o755)
    if workingDirectorySuffix is None:
        # Create unique directory
        workingDirectory = tempfile.mkdtemp(
            prefix=task.__class__.__name__ + "_",
            dir=parentDirectory,
        )
        os.chmod(workingDirectory, 0o755)
        workingDirectory = pathlib.Path(workingDirectory)
    else:
        # Here we assume that the user knows what he is doing and there's no
        # race condition for creating the working directory!
        workingDirectoryName = (
            task.__class__.__name__ + "_" + str(workingDirectorySuffix)
        )
        workingDirectory = parentDirectory / workingDirectoryName
        index = 1
        while workingDirectory.exists():
            workingDirectoryName = (
                task.__class__.__name__
                + "_"
                + str(workingDirectorySuffix)
                + "_{0:02d}".format(index)
            )
            workingDirectory = parentDirectory / workingDirectoryName
            index += 1
        workingDirectory.mkdir(mode=0o775, parents=True, exist_ok=False)
    workingDirectory = stripDataDirectoryPrefix(workingDirectory)
    return workingDirectory


def createPyarchFilePath(file_path):
    """
    This method translates from an ESRF "visitor" path to a "pyarch" path:
    /data/visitor/mx415/id14eh1/20100209 -> /data/pyarch/2010/id14eh1/mx415/20100209
    """
    pyarch_file_path = None
    if isinstance(file_path, str):
        file_path = pathlib.Path(file_path)
    list_of_directories = list(file_path.parts)
    if UtilsConfig.isEMBL():
        if "p13" in list_of_directories[0:3] or "P13" in list_of_directories[0:3]:
            pyarch_file_path = os.path.join("/data/ispyb/p13", *list_of_directories[4:])
        else:
            pyarch_file_path = os.path.join("/data/ispyb/p14", *list_of_directories[4:])
        return pyarch_file_path
    list_beamlines = [
        "bm07",
        "id23eh1",
        "id23eh2",
        "id29",
        "id30a1",
        "id30a2",
        "id30a3",
        "id30b",
    ]

    if (
        "data" in list_of_directories
        and len(list_of_directories) > 5
        and list_of_directories[1] != "data"
    ):
        while list_of_directories[1] != "data" and len(list_of_directories) > 5:
            del list_of_directories[1]

    # Check that we have at least four levels of directories:
    if len(list_of_directories) > 5:
        data_directory = list_of_directories[1]
        second_directory = list_of_directories[2]
        third_directory = list_of_directories[3]
        fourth_directory = list_of_directories[4]
        fifth_directory = list_of_directories[5]
        year = fifth_directory[0:4]
        proposal = None
        beamline = None
        if data_directory == "data" and second_directory == "gz":
            if third_directory == "visitor":
                proposal = fourth_directory
                beamline = fifth_directory
            elif fourth_directory == "inhouse":
                proposal = fifth_directory
                beamline = third_directory
            else:
                raise RuntimeError(
                    "Illegal path for UtilsPath.createPyarchFilePath: "
                    + "{0}".format(file_path)
                )
            list_of_remaining_directories = list_of_directories[6:]
        elif data_directory == "data" and second_directory == "visitor":
            proposal = list_of_directories[3]
            beamline = list_of_directories[4]
            list_of_remaining_directories = list_of_directories[5:]
        elif data_directory == "data" and second_directory in list_beamlines:
            beamline = second_directory
            proposal = list_of_directories[4]
            list_of_remaining_directories = list_of_directories[5:]
        if proposal is not None and beamline is not None:
            pyarch_file_path = pathlib.Path("/data/pyarch") / year / beamline
            pyarch_file_path = pyarch_file_path / proposal
            for directory in list_of_remaining_directories:
                pyarch_file_path = pyarch_file_path / directory
    if pyarch_file_path is None:
        logger.warning(
            "UtilsPath.createPyarchFilePath: path not converted for"
            + " pyarch: %s " % file_path
        )
    return pyarch_file_path


def waitForFile(file, expectedSize=None, timeOut=DEFAULT_TIMEOUT):
    file_path = pathlib.Path(file)
    final_size = None
    has_timed_out = False
    should_continue = True
    file_dir = file_path.parent
    if os.name != "nt" and file_dir.exists():
        # Patch provided by Sebastien 2018/02/09 for forcing NFS cache:
        # logger.debug("NFS cache clear, doing os.fstat on directory {0}".format(fileDir))
        fd = os.open(file_dir.as_posix(), os.O_DIRECTORY)
        stat_result = os.fstat(fd)
        os.close(fd)
        # logger.debug("Results of os.fstat: {0}".format(statResult))
    # Check if file is there
    if file_path.exists():
        file_size = file_path.stat().st_size
        if expectedSize is not None:
            # Check size
            if file_size > expectedSize:
                should_continue = False
        final_size = file_size
    if should_continue:
        logger.info("Waiting for file %s" % file_path)
        #
        time_start = time.time()
        while should_continue and not has_timed_out:
            if os.name != "nt" and file_dir.exists():
                # Patch provided by Sebastien 2018/02/09 for forcing NFS cache:
                # logger.debug("NFS cache clear, doing os.fstat on directory {0}".format(fileDir))
                fd = os.open(file_dir.as_posix(), os.O_DIRECTORY)
                stat_result = os.fstat(fd)  # noqa F841
                os.close(fd)
                # logger.debug("Results of os.fstat: {0}".format(statResult))
            time_elapsed = time.time() - time_start
            # Check if time out
            if time_elapsed > timeOut:
                has_timed_out = True
                str_warning = "Timeout while waiting for file %s" % file_path
                logger.warning(str_warning)
            else:
                # Check if file is there
                if file_path.exists():
                    file_size = file_path.stat().st_size
                    if expectedSize is not None:
                        # Check that it has right size
                        if file_size > expectedSize:
                            should_continue = False
                    else:
                        should_continue = False
                    final_size = file_size
            if should_continue:
                # Sleep 1 s
                time.sleep(1)
    return has_timed_out, final_size


def stripDataDirectoryPrefix(data_directory):
    """ Removes any paths before /data/..., e.g. /gpfs/easy/data/..."""
    list_paths = str(data_directory).split(os.sep)
    if "data" in list_paths:
        while list_paths[1] != "data":
            list_paths = [list_paths[0]] + list_paths[2:]
        new_data_directory = os.sep.join(list_paths)
    else:
        new_data_directory = data_directory
    return pathlib.Path(new_data_directory)

def systemCopyFile(from_path, to_path):
    p = subprocess.Popen(["cp", from_path, to_path])
    p.wait()

def systemRmTree(treePath, ignore_errors=False):
    try:
        if ignore_errors:
            subprocess.check_call(f"rm -rf {treePath}", shell=True)
        else:
            subprocess.check_call(f"rm -r {treePath} 2>&1 > /dev/null", shell=True)
    except subprocess.CalledProcessError:
        if not ignore_errors:
            raise

def systemCopyTree(from_path, to_path, dirs_exists_ok=False):
    if os.path.exists(to_path):
        if dirs_exists_ok:
            systemRmTree(to_path)
        else:
            raise FileExistsError(to_path)
    p = subprocess.Popen(["cp", "-r", from_path, to_path])
    p.wait()

def getBeamlineProposal(directory):
    listDirectory = directory.split(os.sep)
    try:
        if listDirectory[1] == "data":
            if listDirectory[2] == "visitor":
                beamline = listDirectory[4]
                proposal = listDirectory[3]
            else:
                beamline = listDirectory[2]
                proposal = listDirectory[4]
    except:
        beamline = "unknown"
        proposal = userName
    return beamline, proposal

def getIcatBeamline(beamline):
    dict_beamline = {
        "id23eh1": "ID23-1",
        "id23eh2": "ID23-2",
        "id30a1": "ID30A-1",
        "id30a2": "ID30A-2",
        "id30a3": "ID30A-3",
        "id30b": "ID30b",
        "bm07": "BM07"
    }
    return dict_beamline[beamline]


#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

import os
import time
import graypy
import logging
import logging.handlers

from edna2.utils import UtilsConfig


def addGrayLogHandler(logger):
    server = UtilsConfig.get("Logging", "graylog_server")
    port = UtilsConfig.get("Logging", "graylog_port")
    if server is not None and port is not None:
        graylogHandler = graypy.GELFUDPHandler(server, int(port))
        logger.addHandler(graylogHandler)


def addStreamHandler(logger):
    streamHandler = logging.StreamHandler()
    logFileFormat = "%(asctime)s %(levelname)-8s %(message)s"
    formatter = logging.Formatter(logFileFormat)
    streamHandler.setFormatter(formatter)
    logger.addHandler(streamHandler)


def addConfigFileHandler(logger):
    logPath = UtilsConfig.get("Logging", "log_file_path")
    if logPath is not None:
        if "DATE" in logPath:
            logPath = logPath.replace(
                "DATE", time.strftime("%Y-%m-%d", time.localtime(time.time()))
            )
        if not os.path.exists(os.path.dirname(logPath)):
            os.makedirs(os.path.dirname(logPath))
        maxBytes = int(UtilsConfig.get("Logging", "log_file_maxbytes", 1e7))
        backupCount = int(UtilsConfig.get("Logging", "log_file_backupCount", 0))
        fileHandler = logging.handlers.RotatingFileHandler(
            logPath, maxBytes=maxBytes, backupCount=backupCount
        )
        logFileFormat = UtilsConfig.get("Logging", "log_file_format")
        if logFileFormat is None:
            logFileFormat = (
                "%(asctime)s %(module)-20s %(funcName)-15s %(levelname)-8s %(message)s"
            )
        formatter = logging.Formatter(logFileFormat)
        fileHandler.setFormatter(formatter)
        logger.addHandler(fileHandler)

def addFileHandler(logger, log_path):
    if "DATETIME" in log_path:
        log_path = log_path.replace(
            "DATETIME", time.strftime("%Y%m%d-%H%M%S", time.localtime(time.time()))
        )
    if not os.path.exists(os.path.dirname(log_path)):
        os.makedirs(os.path.dirname(log_path))
    maxBytes = int(UtilsConfig.get("Logging", "log_file_maxbytes", 1e7))
    backupCount = int(UtilsConfig.get("Logging", "log_file_backupCount", 0))
    fileHandler = logging.handlers.RotatingFileHandler(
        log_path, maxBytes=maxBytes, backupCount=backupCount
    )
    logFileFormat = UtilsConfig.get("Logging", "log_file_format")
    if logFileFormat is None:
        logFileFormat = (
            "%(asctime)s %(module)-20s %(funcName)-15s %(levelname)-8s %(message)s"
        )
    formatter = logging.Formatter(logFileFormat)
    fileHandler.setFormatter(formatter)
    logger.addHandler(fileHandler)

def setLoggingLevel(logger, level):
    if level is None:
        level = UtilsConfig.get("Logging", "level")
    if level is None:
        level = "INFO"
    if level == "DEBUG":
        loggingLevel = logging.DEBUG
    elif level == "INFO":
        loggingLevel = logging.INFO
    elif level == "WARNING":
        loggingLevel = logging.WARNING
    elif level == "ERROR":
        loggingLevel = logging.ERROR
    elif level == "CRITICAL":
        loggingLevel = logging.CRITICAL
    elif level == "FATAL":
        loggingLevel = logging.FATAL
    else:
        raise RuntimeError('Unknown logging level: "{0}"'.format(level))
    logger.setLevel(loggingLevel)


def getLogger(level=None):
    logger = logging.getLogger("edna2")
    # Check if handlers already added:
    hasGraylogHandler = False
    hasStreamHandler = False
    hasFileHandler = False
    for handler in logger.handlers:
        if isinstance(handler, graypy.GELFUDPHandler):
            hasGraylogHandler = True
        elif isinstance(handler, logging.handlers.RotatingFileHandler):
            hasFileHandler = True
        elif isinstance(handler, logging.StreamHandler):
            hasStreamHandler = True
    if not hasGraylogHandler:
        addGrayLogHandler(logger)
    if not hasStreamHandler:
        addStreamHandler(logger)
    if not hasFileHandler:
        addConfigFileHandler(logger)
    # Set level
    setLoggingLevel(logger, level)
    return logger



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "16/10/2020"


from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ReadImageHeader import ReadImageHeader
from edna2.tasks.ControlDozor import ControlDozor
from edna2.tasks.XDSTasks import XDSIndexing
from edna2.tasks.MosflmTasks import MosflmGeneratePredictionTask


class ControlPredictionTask(AbstractTask):
    """
    This task receives a list of images or data collection ids and
    returns result of indexing
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {  }
        }

    def run(self, inData):
        outData = {}
        xdsIndexingOutData = inData["indexingSolution"]
        # Run MOSFLM prediction
        mosflmUB = xdsIndexingOutData["xparm"]["mosflmUB"]
        mosflmU = xdsIndexingOutData["xparm"]["mosflmU"]
        matrix = {
            "matrixA": mosflmUB,
            "missettingsAngles": [
                0.0,
                0.0,
                0.0
            ],
            "matrixU": mosflmU,
            "cell": {
                "a": xdsIndexingOutData["idxref"]["a"],
                "b": xdsIndexingOutData["idxref"]["b"],
                "c": xdsIndexingOutData["idxref"]["c"],
                "alpha": xdsIndexingOutData["idxref"]["alpha"],
                "beta": xdsIndexingOutData["idxref"]["beta"],
                "gamma": xdsIndexingOutData["idxref"]["gamma"]
            },
        }
        mosflmInData = MosflmGeneratePredictionTask.generateMOSFLMInData(inData={"subWedge": listSubWedge})
        mosflmInData["matrix"] = matrix
        mosflmInData.update({
            "mosaicityEstimation": xdsIndexingOutData["idxref"]["mosaicity"],
            "deviationAngular": 1.0,
            "refinedDistance": xdsIndexingOutData["idxref"]["distance"],
            "symmetry": "P1"
        })
        mosflmGeneratePredictionTask =MosflmGeneratePredictionTask(inData=mosflmInData)
        mosflmGeneratePredictionTask.execute()
        pass
        return outData




#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "26/07/2019"

import pathlib

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.H5ToCBFTask import H5ToCBFTask

class DistlSignalStrengthTask(AbstractTask):
    """
    This task runs the labelit.distl command for pre-screening reference images.
    """

    def run(self, inData):
        commandLine = 'distl.signal_strength '
        image_path = pathlib.Path(inData['referenceImage'])
        if image_path.suffix == ".h5":
            inDataH5ToCBF = {
                "imageNumber": 1,
                "hdf5File": str(image_path)
            }
            h5ToCBFTask = H5ToCBFTask(inData=inDataH5ToCBF)
            h5ToCBFTask.execute()
            image_path = h5ToCBFTask.outData["outputCBFFile"]
        commandLine += str(image_path)
        logPath = self.getWorkingDirectory() / 'distl.log'
        self.runCommandLine(commandLine, logPath=logPath)
        with open(str(logPath)) as f:
            logText = f.read()
        imageQualityIndicators = self.parseLabelitDistlOutput(logText)
        outData = {'imageQualityIndicators': imageQualityIndicators}
        return outData

    def parseLabelitDistlOutput(self, logText):
        imageQualityIndicators = {}
        for line in logText.split("\n"):
            if line.find("Spot Total") != -1:
                spotTotal = int(line.split()[3])
                imageQualityIndicators['spotTotal'] = spotTotal
            elif line.find("In-Resolution Total") != -1:
                inResTotal = int(line.split()[3])
                imageQualityIndicators['inResTotal'] = inResTotal
            elif line.find("Good Bragg Candidates") != -1:
                goodBraggCandidates = int(line.split()[4])
                imageQualityIndicators['goodBraggCandidates'] = goodBraggCandidates
            elif line.find("Ice Rings") != -1:
                iceRings = int(line.split()[3])
                imageQualityIndicators['iceRings'] = iceRings
            elif line.find("Method 1 Resolution") != -1:
                method1Res = float(line.split()[4])
                imageQualityIndicators['method1Res'] = method1Res
            elif line.find("Method 2 Resolution") != -1:
                if line.split()[4] != "None":
                    fMethod2Res = float(line.split()[4])
                    imageQualityIndicators['method2Res'] = fMethod2Res
            elif line.find("Maximum unit cell") != -1:
                if line.split()[4] != "None":
                    fMaxUnitCell = float(line.split()[4])
                    imageQualityIndicators['maxUnitCell'] = fMaxUnitCell
            elif line.find("%Saturation, Top 50 Peaks") != -1:
                pctSaturationTop50Peaks = float(line.split()[5])
                imageQualityIndicators['pctSaturationTop50Peaks'] = pctSaturationTop50Peaks
            elif line.find("In-Resolution Ovrld Spots") != -1:
                inResolutionOvrlSpots = int(line.split()[4])
                imageQualityIndicators['inResolutionOvrlSpots'] = inResolutionOvrlSpots
            elif line.find("Bin population cutoff for method 2 resolution") != -1:
                binPopCutOffMethod2Res = float(line.split()[7][:-1])
                imageQualityIndicators['binPopCutOffMethod2Res'] = binPopCutOffMethod2Res
            elif line.find("Total integrated signal, pixel-ADC units above local background (just the good Bragg candidates)") != -1:
                totalIntegratedSignal = float(line.split()[-1])
                imageQualityIndicators['totalIntegratedSignal'] = totalIntegratedSignal
            elif line.find("signals range from") != -1:
                listStrLine = line.split()
                signalRangeMin = float(listStrLine[3])
                signalRangeMax = float(listStrLine[5])
                signalRangeAverage = float(listStrLine[-1])
                imageQualityIndicators['signalRangeMin'] = signalRangeMin
                imageQualityIndicators['signalRangeMax'] = signalRangeMax
                imageQualityIndicators['signalRangeAverage'] = signalRangeAverage
            elif line.find("Saturations range from") != -1:
                listStrLine = line.split()
                saturationRangeMin = float(listStrLine[3][:-1])
                saturationRangeMax = float(listStrLine[5][:-1])
                saturationRangeAverage = float(listStrLine[-1][:-1])
                imageQualityIndicators['saturationRangeMin'] = saturationRangeMin
                imageQualityIndicators['saturationRangeMax'] = saturationRangeMax
                imageQualityIndicators['saturationRangeAverage'] = saturationRangeAverage
        return imageQualityIndicators



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# kernel/src/EDUtilsImage.py

import os
import re
import fabio
import pathlib


def __compileAndMatchRegexpTemplate(pathToImage):
    listResult = []
    if not isinstance(pathToImage, pathlib.Path):
        pathToImage = pathlib.Path(str(pathToImage))
    baseImageName = pathToImage.name
    regexp = re.compile(r"(.*)([^0^1^2^3^4^5^6^7^8^9])([0-9]*)\.(.*)")
    match = regexp.match(baseImageName)
    if match is not None:
        listResult = [
            match.group(0),
            match.group(1),
            match.group(2),
            match.group(3),
            match.group(4),
        ]
    return listResult


def getImageNumber(pathToImage):
    iImageNumber = None
    listResult = __compileAndMatchRegexpTemplate(pathToImage)
    if listResult is not None:
        iImageNumber = int(listResult[3])
    return iImageNumber


def getTemplate(pathToImage, symbol="#"):
    template = None
    listResult = __compileAndMatchRegexpTemplate(pathToImage)
    if listResult is not None:
        prefix = listResult[1]
        separator = listResult[2]
        imageNumber = listResult[3]
        suffix = listResult[4]
        hashes = ""
        for i in range(len(imageNumber)):
            hashes += symbol
        template = prefix + separator + hashes + "." + suffix
    return template


def getPrefix(pathToImage):
    prefix = None
    listResult = __compileAndMatchRegexpTemplate(pathToImage)
    if listResult is not None:
        prefix = listResult[1]
    return prefix


def getSuffix(pathToImage):
    suffix = None
    listResult = __compileAndMatchRegexpTemplate(pathToImage)
    if listResult is not None:
        suffix = listResult[4]
    return suffix


def getPrefixNumber(pathToImage):
    prefix = getPrefix(pathToImage)
    number = getImageNumber(pathToImage)
    prefixNumber = "{0}_{1:04d}".format(prefix, number)
    return prefixNumber


def splitPrefixRunNumber(path_to_image):
    file_name = pathlib.Path(path_to_image).name
    list_parts = str(file_name).split("_")
    pre_prefix = "_".join(list_parts[0:-2])
    run_number = int(list_parts[-2])
    return pre_prefix, run_number


def getH5FilePath(filePath, batchSize=100, hasOverlap=False, isFastMesh=False):
    if type(filePath) == str:
        filePath = pathlib.Path(filePath)
    imageNumber = getImageNumber(filePath)
    prefix = getPrefix(filePath)
    if hasOverlap or filePath.name.startswith("ref-"):
        h5ImageNumber = 1
        h5FileNumber = imageNumber
    elif (
        isFastMesh
        or filePath.name.startswith("mesh-")
        or filePath.name.startswith("line-")
    ):
        h5ImageNumber = int((imageNumber - 1) / 100) + 1
        h5FileNumber = 1
    else:
        h5ImageNumber = 1
        h5FileNumber = int((imageNumber - 1) / batchSize) * batchSize + 1
    h5MasterFileName = "{prefix}_{h5FileNumber}_master.h5".format(
        prefix=prefix, h5FileNumber=h5FileNumber
    )
    h5MasterFilePath = filePath.parent / h5MasterFileName
    h5DataFileName = "{prefix}_{h5FileNumber}_data_{h5ImageNumber:06d}.h5".format(
        prefix=prefix, h5FileNumber=h5FileNumber, h5ImageNumber=h5ImageNumber
    )
    h5DataFilePath = filePath.parent / h5DataFileName
    return h5MasterFilePath, h5DataFilePath, h5FileNumber


def mergeCbf(listPath, outputPath):
    firstImage = None
    no_images = len(listPath)
    for imagePath in listPath:
        if firstImage is None:
            firstImage = fabio.open(imagePath)
        else:
            image = fabio.open(imagePath)
            firstImage.data += image.data
    header_contents = firstImage.header["_array_data.header_contents"]
    list_header_contents = header_contents.split("\n")
    index = 0
    done = False
    while not done:
        line = list_header_contents[index]
        if line.startswith("# Angle_increment"):
            list_angle_increment = line.split(" ")
            angle_increment = float(list_angle_increment[2])
            new_angle_increment = angle_increment * no_images
            list_angle_increment[2] = str(new_angle_increment)
            new_line = " ".join(list_angle_increment)
            list_header_contents[index] = new_line
            done = True
        else:
            index += 1
    new_header_contents = "\n".join(list_header_contents)
    firstImage.header["_array_data.header_contents"] = new_header_contents
    firstImage.write(outputPath)
    return


def mergeCbfInDirectory(cbfDirectory, prefix=None, newPrefix=None):
    path_to_dir = pathlib.Path(cbfDirectory)
    index = None
    list_dir = [str(path) for path in list(path_to_dir.glob("*.cbf"))]
    list_dir.sort()
    list_of_image_lists = []
    listImage = None
    for cbf_file in list_dir:
        if not os.path.basename(cbf_file).startswith("ref-"):
            image_no = getImageNumber(cbf_file)
            if index is None or index != image_no:
                print("Starting image: {0}".format(cbf_file))
                index = image_no
                if listImage is not None:
                    list_of_image_lists.append(listImage)
                listImage = [cbf_file]
            else:
                print("Adding image no {0}".format(image_no))
                listImage.append(cbf_file)
            index += 1
    list_of_image_lists.append(listImage)
    image_number = 1
    firstImage = list_of_image_lists[0][0]
    directory = os.path.dirname(firstImage)
    old_run_number = None
    for list_image in list_of_image_lists:
        pre_prefix, run_number = splitPrefixRunNumber(list_image[0])
        new_run_Number = run_number + 10
        if old_run_number is None or old_run_number != new_run_Number:
            old_run_number = new_run_Number
            image_number = 1
        new_cbf_path = "{0}/{1}_{2}_{3:04d}.cbf".format(
            directory, pre_prefix, new_run_Number, image_number
        )
        mergeCbf(list_image, new_cbf_path)
        image_number += 1


def crop_image(image_path):
    #
    # Crop image
    #
    os.system(f"convert -trim -border 5 -bordercolor white {image_path} {image_path}")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# kernel/src/EDUtilsTable.py

import json
import xmltodict

from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


def getDict(dnaTablesPath):
    with open(str(dnaTablesPath)) as f:
        dnaTables = f.read()
    if "</dna_tables>" not in dnaTables:
        # Fix for bug in MOSFLM
        dnaTables += "</dna_tables>"
    orderedDictDnaTables = xmltodict.parse(dnaTables)
    dictDnaTables = json.loads(json.dumps(orderedDictDnaTables))
    return dictDnaTables


def getTables(dictDnaTables, tableName):
    listTables = []
    listTable = dictDnaTables["dna_tables"]["table"]
    for table in listTable:
        if tableName == table["@name"]:
            listTables.append(table)
    return listTables


def getListParam(table):
    if isinstance(table["list"], list):
        listParam = table["list"]
    else:
        listParam = [table["list"]]
    return listParam


def getItemValue(dictParameter, key):
    value = None
    if isinstance(dictParameter["item"], list):
        listItem = dictParameter["item"]
    else:
        listItem = [dictParameter["item"]]
    for item in listItem:
        if item["@name"] == key:
            value = item["#text"]
    return _convertFromString(value)


def _convertFromString(value):
    if value is not None:
        try:
            if "." in value:
                value = float(value)
            else:
                value = int(value)
        except ValueError:
            # The value is returned as a string...
            pass
    return value


def getListValue(listParameter, key1, key2):
    value = None
    for dictParameter in listParameter:
        if dictParameter["@name"] == key1:
            if isinstance(dictParameter["item"], list):
                for item in dictParameter["item"]:
                    if item["@name"] == key2:
                        value = item["#text"]
            else:
                value = dictParameter["item"]["#text"]
    return _convertFromString(value)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the 'Software'), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "26/03/2020"

DETECTOR_PARAMS = {
    "pilatus2m": {
        "nx": 1475,
        "ny": 1679,
        "pixel": 0.172,
        "sensorThickness": 0.32,
        "xdsUntrustedRectangle": [
            [487, 495, 0, 1680],
            [981, 989, 0, 1680],
            [0, 1476, 195, 213],
            [0, 1476, 407, 425],
            [0, 1476, 619, 637],
            [0, 1476, 831, 849],
            [0, 1476, 1043, 1061],
            [0, 1476, 1255, 1273],
            [0, 1476, 1467, 1485],
        ],
    },
    "pilatus6m": {
        "nx": 2463,
        "ny": 2527,
        "pixel": 0.172,
        "sensorThickness": 0.32,
        "xdsUntrustedRectangle": [
            [487, 495, 0, 2528],
            [981, 989, 0, 2528],
            [1475, 1483, 0, 2528],
            [1969, 1977, 0, 2528],
            [0, 2464, 195, 213],
            [0, 2464, 407, 425],
            [0, 2464, 619, 637],
            [0, 2464, 831, 849],
            [0, 2464, 1043, 1061],
            [0, 2464, 1255, 1273],
            [0, 2464, 1467, 1485],
            [0, 2464, 1679, 1697],
            [0, 2464, 1891, 1909],
            [0, 2464, 2103, 2121],
            [0, 2464, 2315, 2333],
        ],
    },
    "eiger4m": {
        "nx": 2070,
        "ny": 2167,
        "pixel": 0.075,
        "sensorThickness": 0.32,
        "xdsUntrustedRectangle": [
            [1029, 1040, 0, 2167],
            [0, 2070, 512, 550],
            [0, 2070, 1063, 1103],
            [0, 2070, 1614, 1654],
        ],
    },
    "eiger9m": {
        "nx": 3108,
        "ny": 3262,
        "pixel": 0.075,
        "sensorThickness": 0.75,
        "xdsUntrustedRectangle": [],
    },
    "eiger16m": {
        "nx": 4148,
        "ny": 4362,
        "pixel": 0.075,
        "sensorThickness": 0.75,
        "xdsUntrustedRectangle": [],
    },
}


def __getDetectorValue(detectorType, key):
    if detectorType in DETECTOR_PARAMS:
        p = DETECTOR_PARAMS[detectorType]
        return p[key]
    else:
        raise RuntimeError(
            'Detector type "{0}" not defined in UtilsDetector.py!'.format(detectorType)
        )


def getNx(detectorType):
    return __getDetectorValue(detectorType, "nx")


def getNy(detectorType):
    return __getDetectorValue(detectorType, "ny")


def getPixelsize(detectorType):
    return __getDetectorValue(detectorType, "pixel")


def getXdsUntrustedRectangle(detectorType):
    return __getDetectorValue(detectorType, "xdsUntrustedRectangle")


def getSensorThickness(detectorType):
    return __getDetectorValue(detectorType, "sensorThickness")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxPluginExec/plugins/EDPluginMXWaitFile-v1.1/plugins/EDPluginMXWaitFilev1_1.py

import pathlib

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsPath
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()

DEFAULT_TIMEOUT = 120  # s


class WaitFileTask(AbstractTask):
    """
    This plugin waits for a file to appear on disk
    """

    def run(self, inData):
        # Wait for file if it's not already on disk'
        if "file" not in inData:
            raise BaseException("No expected file path in input!")
        filePath = pathlib.Path(self.inData["file"])
        expectedSize = inData.get("expectedSize", None)
        configTimeOut = UtilsConfig.get(self, "timeOut", DEFAULT_TIMEOUT)
        timeOut = inData.get("timeOut", configTimeOut)
        hasTimedOut, finalSize = UtilsPath.waitForFile(
            filePath, expectedSize=expectedSize, timeOut=timeOut
        )
        outData = {"timedOut": hasTimedOut}
        if finalSize is not None:
            outData["finalSize"] = finalSize
        return outData



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "11/01/2021"



from edna2.tasks.AbstractTask import AbstractTask




class Raddose(AbstractTask):
    """
    This task runs the program raddose
    """
    HENDERSON_LIMIT = 2e7

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "exposureTime": {"type": "number"},
                "flux": {"type": "number"},
                "beamSizeX": {"type": "number"},
                "beamSizeY": {"type": "number"},
                "wavelength": {"type": "number"},
                "numberOfImages": {"type": "number"},
            }
        }

    def run(self, inData):
        commandLine, listCommand = self.createCommandLine(inData)
        self.runCommandLine(commandLine, listCommand=listCommand)
        dictResults = self.parseLogFile(self.getLogPath())
        outData = self.createOutData(inData, dictResults, pathToLogFile=self.getLogPath())
        return outData

    @staticmethod
    def createCommandLine(inData):
        commandLine = "raddose"
        experimentalCondition = inData["experimentalCondition"]
        sample = inData["sample"]
        beam = experimentalCondition["beam"]
        numOperators = inData["numOperators"]
        # Analyse chemical composition
        chemicalComposition = inData["chemicalComposition"]
        # Solvent
        totalSATM = []
        solvent = chemicalComposition.get("solvent")
        if solvent is not None:
            totalSATM = solvent["atom"]
        # Structure
        structure = chemicalComposition.get("structure", None)
        totalNRESInStructure = 0
        totalNDNAInStructure = 0
        totalNRNAInStructure = 0
        totalPATM = []
        if structure is not None:
            listChain = structure["chain"]

            for chain in listChain:
                # heavy atoms of each chain to be added in the PATM
                atomicCompositionHeavyAtoms = chain.get("heavyAtoms", None)
                if atomicCompositionHeavyAtoms is not None:
                    iterator = 1
                    while iterator <= chain["numberOfCopies"]:
                        totalPATM = Raddose.mergeAtomicComposition(totalPATM, atomicCompositionHeavyAtoms)
                        iterator = iterator + 1

                type = chain["type"]
                numberOfMonomers = chain["numberOfMonomers"] * chain["numberOfCopies"]

                if (type == "protein"):
                    totalNRESInStructure = totalNRESInStructure + numberOfMonomers
                elif(type == "dna"):
                    totalNDNAInStructure = totalNDNAInStructure + numberOfMonomers
                elif(type == "rna"):
                    totalNRNAInStructure = totalNRNAInStructure + numberOfMonomers

            numStructInAU = structure["numberOfCopiesInAsymmetricUnit"]
            numStructInUC = int(numStructInAU * numOperators)

            # Ligands - not implemented yet
            if "ligand" in structure:
                raise RuntimeError("Ligand present in structure but support for ligands not yet implemented!")
            # xsDataLigands = xsDataStructure.getLigand()
            # for ligand in xsDataLigands:
            #
            #     # Light atoms to be added to the NRES
            #     nres = ligand.getNumberOfLightAtoms().getValue() * ligand.getNumberOfCopies().getValue() / 7.85
            #     totalNRESInStructure = totalNRESInStructure + nres
            #
            #     # Heavy atoms to be added to the PATM
            #     if (ligand.getHeavyAtoms() is not None):
            #         iterator = 1
            #         while iterator <= ligand.getNumberOfCopies().getValue():
            #             totalPATM = self.mergeAtomicComposition(totalPATM, ligand.getHeavyAtoms())
            #             iterator = iterator + 1

        listCommand = [
            "BEAM {x} {y}".format(**beam["size"]),
            "PHOSEC {flux}".format(**beam),
            "WAVELENGTH {wavelength}".format(**beam),
            # "CRYSTAL {x} {y} {z}".format(**sample["size"]),
            "CELL {a} {b} {c} {alpha} {beta} {gamma}".format(**inData["cell"]),
            "EXPOSURE {exposureTime}".format(**beam),
            "IMAGES {numberOfImages}".format(**inData),
            "NMON {0}".format(numStructInUC)
        ]

        if totalNRESInStructure != 0:
            listCommand.append("NRES {0}".format(int(round(totalNRESInStructure))))
        if totalNDNAInStructure != 0:
            listCommand.append("NDNA {0}".format(int(round(totalNDNAInStructure))))
        if totalNRNAInStructure != 0:
            listCommand.append("NRA {0}".format(int(round(totalNRNAInStructure))))
        if len(totalPATM) != 0:
            patmLine = "PATM"
            for patm in totalPATM:
                patmLine += " {symbol} {numberOf}".format(**patm)
            listCommand.append(patmLine)
        if len(totalSATM) != 0:
            satmLine = "SATM"
            for satm in totalSATM:
                satmLine += " {symbol} {concentration}".format(**satm)
            listCommand.append(satmLine)

        #
        # patmLine = "PATM"
        # for patm in inData["crystalPATM"]:
        #     patmLine += " {symbol} {numberOf}".format(**patm)
        # listCommand.append(patmLine)
        listCommand.append("END")
        return commandLine, listCommand

    @staticmethod
    def parseLogFile(logFilePath):
        dictResults = {}
        with open(str(logFilePath), "rb") as fd:
            content = fd.read()
        # Cut off the last 150 bytes
        content = content[0:len(content) - 150]
        for line in content.decode('utf-8').split("\n"):
            if "Dose in Grays" in line:
                dictResults["doseInGrays"] = round(float(line.split()[-1]), 1)
            elif "Total absorbed dose (Gy)" in line:
                dictResults["totalAbsorbedDose"] = round(float(line.split()[-1]), 1)
            elif "Solvent Content (%)" in line:
                dictResults["solventContent"] = round(float(line.split()[-1]), 1)
        return dictResults

    @staticmethod
    def createOutData(inData, dictResults, pathToLogFile=None):
        outData = {}
        numberOfImages = inData["numberOfImages"]
        exposureTime = inData["experimentalCondition"]["beam"]["exposureTime"]
        totalExposureTime = numberOfImages * exposureTime
        if "doseInGrays" in dictResults:
            absorbedDose = dictResults["doseInGrays"]
        elif "totalAbsorbedDose" in dictResults:
            absorbedDose = dictResults["totalAbsorbedDose"]
        else:
            raise RuntimeError["Neither doseInGrays nor totalAbsorbedDose in results!"]
        absorbedDoseRate = round(absorbedDose / totalExposureTime, 1)
        timeToReachHendersonLimit = round(Raddose.HENDERSON_LIMIT / absorbedDoseRate, 1)
        outData = {
            "absorbedDose": absorbedDose,
            "absorbedDoseRate": absorbedDoseRate,
            "pathToLogFile": pathToLogFile,
            "timeToReachHendersonLimit": timeToReachHendersonLimit
        }
        return outData

    @staticmethod
    def mergeAtomicComposition(atomicComposition1, atomicComposition2):

        mergedAtomicComposition = []
        dictionary = {}

        for atom2 in atomicComposition2:
            dictionary[atom2["symbol"]] = atom2["numberOf"]

        for atom1 in atomicComposition1:
            symbol = atom1["symbol"]
            if Raddose.exists(symbol, atomicComposition2):
                mergedAtom = {
                    "symbol": symbol,
                    "numberOf": atom1["numberOf"] + dictionary[symbol]
                }
                mergedAtomicComposition.append(mergedAtom)
            else:
                mergedAtomicComposition.append(atom1)

        for atom2 in atomicComposition2:
            symbol = atom2["symbol"]
            if not Raddose.exists(symbol, atomicComposition1):
                mergedAtomicComposition.append(atom2)

        return mergedAtomicComposition

    @staticmethod
    def exists(symbol, atomicComposition):
        exists = False
        for atom in atomicComposition:
            if atom["symbol"] == symbol:
                exists = True
                break
        return exists



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__author__ = "Olof Svensson"
__contact__ = "svensson@esrf.eu"
__copyright__ = "ESRF"
__updated__ = "2022-02-13"

import os
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsLogging


logger = UtilsLogging.getLogger()


class DozorRD(AbstractTask):  # pylint: disable=too-many-instance-attributes
    """
    The DozorM2 is responsible for executing the 'dozorm2' program.
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "list_dozor_all": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "wavelength": {"type": "number"},
                "exposureTime": {"type": "number"},
                "numberOfImages": {"type": "number"},
            },
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "logPath": {"type": "string"},
                "noSpotDecrease": {"type": "number"},
                "mainScore": {"type": "number"},
                "spotIntensity": {"type": "number"},
                "sumIntensity": {"type": "number"},
                "average": {"type": "number"},
            },
        }

    def run(self, inData):
        outData = {}
        commandLine = "dozorrd dozorrd.dat"
        for index, dozor_all_file in enumerate(inData["list_dozor_all"]):
            os.symlink(
                dozor_all_file, str(self.getWorkingDirectory() / "d_{0:03d}".format(index + 1))
            )
        commands = self.generateCommands(inData)
        with open(str(self.getWorkingDirectory() / "dozorrd.dat"), "w") as f:
            f.write(commands)
        self.setLogFileName("dozorrd.log")
        self.runCommandLine(commandLine)
        # Send log and error log to the logger
        logger.info("\n\n" + self.getLog())
        logger.info("")
        errorLog = self.getErrorLog()
        if errorLog != "":
            logger.info("\n\n" + self.getErrorLog())
            logger.info("")
        outData = self.parseDozorRDLogFile(self.getLogPath())
        # Generate plots
        plotMtvFile = self.getWorkingDirectory() / "dozor_rd.mtv"
        if plotMtvFile.exists():
            listPlotFile = self.generatePngPlots(
                plotmtvFile=plotMtvFile,
                workingDir=self.getWorkingDirectory()
            )
            outData["listPlotFile"] = listPlotFile
        return outData

    @staticmethod
    def generateCommands(inData):
        """
        This method creates the input file for dozorm
        """
        nameTemplateScan = "d_0??"
        command = "X-ray_wavelength {0}\n".format(inData["wavelength"])
        command += "exposure {0}\n".format(inData["exposureTime"])
        command += "number_images {0}\n".format(inData["numberOfImages"])
        command += "first_scan_number 1\n"
        command += "number_scans {0}\n".format(len(inData["list_dozor_all"]))
        command += "name_template_scan {0}\n".format(nameTemplateScan)
        command += "end\n"
        return command

    @staticmethod
    def parseDozorRDLogFile(logPath):
        #  Program dozorrd /A.Popov,G.Bourenkov /
        #  Version 1.1 //  31.01.2022
        #  Copyright 2020 by Alexander Popov and Gleb Bourenkov
        #   T1/2 estimates
        #  N.of spot decrease= 1.751
        #       main score   =10.719
        #   spot intensity   =67.269
        #   sum  intensity   = 2.218
        #   average   =26.735
        outData = {}
        outData["logPath"] = str(logPath)
        with open(str(logPath)) as fd:
            listLogLines = fd.readlines()
        dictParse = {
            "N.of spot decrease": "noSpotDecrease",
            "main score": "mainScore",
            "spot intensity": "spotIntensity",
            "sum  intensity": "sumIntensity",
            "average": "average",
        }
        for line in listLogLines:
            if "=" in line:
                key, value = line.split("=")
                key = key.strip()
                if key in dictParse:
                    try:
                        outData[dictParse[key]] = float(value)
                    except ValueError:
                        outData[dictParse[key]] = -9999
                        errorMessage = "Error when parsing '{0}': '{1}'".format(key, value.strip())
                        if "errorMessage" in outData:
                            outData["errorMessage"] += " " + errorMessage
                        else:
                            outData["errorMessage"] = errorMessage
        return outData

    def generatePngPlots(self, plotmtvFile, workingDir):
        listPlotFile = []
        # Create plot dictionary
        with open(plotmtvFile) as f:
            listLines = f.readlines()
        dictPlot = None
        plotData = None
        dictPlotList = None
        listPlots = []
        index = 0
        while index < len(listLines):
            # print("0" + listLines[index])
            if listLines[index].startswith("$"):
                dictPlot = {}
                dictPlotList = []
                listPlots.append(dictPlot)
                dictPlot["plotList"] = dictPlotList
                index += 1
                dictPlot["name"] = listLines[index].split("'")[1]
                index += 1
                # print(listLines[index])
                while listLines[index].startswith("%"):
                    listLine = listLines[index].split("=")
                    # print(listLine)
                    label = listLine[0][1:].strip()
                    # print("label: " + str([label]))
                    if "'" in listLine[1]:
                        value = listLine[1].split("'")[1]
                    else:
                        value = listLine[1]
                    value = value.replace("\n", "").strip()
                    # print("value: " + str([value]))
                    dictPlot[label] = value
                    index += 1
                    # print(listLines[index])
            elif listLines[index].startswith("#"):
                dictSubPlot = {}
                dictPlotList.append(dictSubPlot)
                plotName = listLines[index].split("#")[1].replace("\n", "").strip()
                dictSubPlot["name"] = plotName
                index += 1
                # print("1" + listLines[index])
                while listLines[index].startswith("%"):
                    listLine = listLines[index].split("=")
                    # print(listLine)
                    label = listLine[0][1:].strip()
                    # print("label: " + str([label]))
                    if "'" in listLine[1]:
                        value = listLine[1].split("'")[1]
                    else:
                        value = listLine[1]
                    value = value.replace("\n", "").strip()
                    # print("value: " + str([value]))
                    dictSubPlot[label] = value
                    index += 1
                    # print(listLines[index])
                dictSubPlot["xValues"] = []
                dictSubPlot["yValues"] = []
            else:
                listData = listLines[index].replace("\n", "").split()
                dictSubPlot["xValues"].append(float(listData[0]))
                dictSubPlot["yValues"].append(float(listData[1]))
                index += 1
        # pprint.pprint(listPlots)
        # Generate the plots
        for mtvplot in listPlots:
            listLegend = []
            xmin = None
            xmax = None
            ymin = None
            ymax = None
            for subPlot in mtvplot["plotList"]:
                xmin1 = min(subPlot["xValues"])
                if xmin is None or xmin > xmin1:
                    xmin = xmin1
                xmax1 = max(subPlot["xValues"])
                if xmax is None or xmax < xmax1:
                    xmax = xmax1
                ymin1 = min(subPlot["yValues"])
                if ymin is None or ymin > ymin1:
                    ymin = ymin1
                ymax1 = max(subPlot["yValues"])
                if ymax is None or ymax < ymax1:
                    ymax = ymax1
            if "xmin" in mtvplot:
                xmin = float(mtvplot["xmin"])
            if "ymin" in mtvplot:
                ymin = float(mtvplot["ymin"])
            plt.xlim(xmin, xmax)
            plt.ylim(ymin, ymax)
            plt.xlabel(mtvplot["xlabel"])
            plt.ylabel(mtvplot["ylabel"])
            plt.title(mtvplot["name"])
            for subPlot in mtvplot["plotList"]:
                if "markercolor" in subPlot:
                    style = "bs-."
                else:
                    style = "r"
                plt.plot(subPlot["xValues"], subPlot["yValues"], style, linewidth=2)
                listLegend.append(subPlot["linelabel"])
            plt.legend(listLegend, loc='lower right')
            plotPath = os.path.join(str(workingDir), mtvplot["name"].replace(" ", "").replace(".", "_") + ".png")
            plt.savefig(plotPath, bbox_inches='tight', dpi=75)
            plt.close()
            listPlotFile.append(plotPath)
        return listPlotFile
